from __future__ import annotations

from allotropy.allotrope.models.adm.pcr.benchling._2023._09.dpcr import Model
from allotropy.allotrope.pandas_util import read_csv
from allotropy.allotrope.schema_mappers.adm.pcr.BENCHLING._2023._09.dpcr import Mapper
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.appbio_absolute_q.appbio_absolute_q_structure import create_data
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser


class AppbioAbsoluteQParser(VendorParser):
    @property
    def display_name(self) -> str:
        return "AppBio AbsoluteQ"

    @property
    def release_state(self) -> ReleaseState:
        return ReleaseState.RECOMMENDED

    def to_allotrope(self, named_file_contents: NamedFileContents) -> Model:
        data_frame = read_csv(named_file_contents.contents)
        data = create_data(data_frame, named_file_contents.original_file_name)
        return self._get_mapper(Mapper).map_model(data)



from __future__ import annotations

import re

from allotropy.allotrope.models.shared.definitions.definitions import NaN
from allotropy.allotrope.schema_mappers.adm.plate_reader.benchling._2023._09.plate_reader import (
    Data,
    ImageFeature,
    Measurement,
    MeasurementGroup,
    MeasurementType,
    Metadata,
    ProcessedData,
)
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.mabtech_apex.mabtech_apex_contents import MabtechApexContents
from allotropy.parsers.utils.pandas import map_rows, SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import assert_not_none

IMAGE_FEATURES = [
    "Spot Forming Units (SFU)",
    "Average Relative Spot Volume (RSV)",
    "Sum of Spot Volume (RSV)",
]


def _create_metadata(contents: MabtechApexContents, file_name: str) -> Metadata:
    machine_id = assert_not_none(
        re.match(
            "([A-Z]+[a-z]+) ([0-9]+)",
            contents.plate_info[str, "Machine ID:"],
        ),
        msg="Unable to interpret Machine ID",
    )

    return Metadata(
        device_identifier=NOT_APPLICABLE,
        device_type="imager",
        detection_type="optical-imaging",
        software_name="Apex",
        unc_path=contents.plate_info.get(str, "Path:"),
        software_version=contents.plate_info.get(str, "Software Version:"),
        model_number=machine_id.group(1),
        equipment_serial_number=machine_id.group(2),
        file_name=file_name,
        analyst=contents.plate_info.get(str, "Saved By:"),
    )


def _create_measurement(plate_data: SeriesData) -> Measurement:
    location_id = plate_data[str, "Well"]
    well_plate = plate_data.get(str, "Plate")

    return Measurement(
        type_=MeasurementType.OPTICAL_IMAGING,
        identifier=random_uuid_str(),
        measurement_time=plate_data[str, "Read Date"],
        location_identifier=location_id,
        well_plate_identifier=well_plate,
        sample_identifier=f"{well_plate}_{location_id}",
        exposure_duration_setting=plate_data.get(float, "Exposure"),
        illumination_setting=plate_data.get(float, "Preset Intensity"),
        processed_data=ProcessedData(
            identifier=random_uuid_str(),
            features=[
                ImageFeature(
                    identifier=random_uuid_str(),
                    feature=feature,
                    result=plate_data.get(float, feature, NaN),
                )
                for feature in IMAGE_FEATURES
            ],
        ),
    )


def _create_measurement_group(data: SeriesData) -> MeasurementGroup:
    return MeasurementGroup(
        measurements=[_create_measurement(data)],
        plate_well_count=96,
    )


def create_data(contents: MabtechApexContents, file_name: str) -> Data:
    # if Read Date is not present in file, return None, no measurement for given Well
    plate_data = contents.data.dropna(subset="Read Date")
    return Data(
        _create_metadata(contents, file_name),
        map_rows(plate_data, _create_measurement_group),
    )



from collections.abc import Callable
from dataclasses import dataclass
from enum import Enum

from allotropy.allotrope.models.adm.plate_reader.benchling._2023._09.plate_reader import (
    CalculatedDataAggregateDocument,
    CalculatedDataDocumentItem,
    ContainerType,
    DataSourceAggregateDocument,
    DataSourceDocumentItem,
    DataSystemDocument,
    DeviceSystemDocument,
    FluorescencePointDetectionDeviceControlAggregateDocument,
    FluorescencePointDetectionDeviceControlDocumentItem,
    FluorescencePointDetectionMeasurementDocumentItems,
    ImageFeatureAggregateDocument,
    ImageFeatureDocumentItem,
    LuminescencePointDetectionDeviceControlAggregateDocument,
    LuminescencePointDetectionDeviceControlDocumentItem,
    LuminescencePointDetectionMeasurementDocumentItems,
    MeasurementAggregateDocument,
    Model,
    OpticalImagingDeviceControlAggregateDocument,
    OpticalImagingDeviceControlDocumentItem,
    OpticalImagingMeasurementDocumentItems,
    PlateReaderAggregateDocument,
    PlateReaderDocumentItem,
    ProcessedDataAggregateDocument,
    ProcessedDataDocumentItem,
    SampleDocument,
    ScanPositionSettingPlateReader,
    TransmittedLightSetting,
    UltravioletAbsorbancePointDetectionDeviceControlAggregateDocument,
    UltravioletAbsorbancePointDetectionDeviceControlDocumentItem,
    UltravioletAbsorbancePointDetectionMeasurementDocumentItems,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueDegreeCelsius,
    TQuantityValueMilliAbsorbanceUnit,
    TQuantityValueMillimeter,
    TQuantityValueMilliSecond,
    TQuantityValueNanometer,
    TQuantityValueNumber,
    TQuantityValueRelativeFluorescenceUnit,
    TQuantityValueRelativeLightUnit,
    TQuantityValueUnitless,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    InvalidJsonFloat,
    JsonFloat,
    TDateTimeValue,
    TQuantityValue,
)
from allotropy.constants import ASM_CONVERTER_VERSION
from allotropy.exceptions import AllotropyParserError
from allotropy.parsers.utils.values import assert_not_none, quantity_or_none


class MeasurementType(Enum):
    OPTICAL_IMAGING = "OPTICAL_IMAGING"
    ULTRAVIOLET_ABSORBANCE = "ULTRAVIOLET_ABSORBANCE"
    FLUORESCENCE = "FLUORESCENCE"
    LUMINESCENCE = "LUMINESCENCE"


MeasurementDocumentItems = (
    UltravioletAbsorbancePointDetectionMeasurementDocumentItems
    | FluorescencePointDetectionMeasurementDocumentItems
    | LuminescencePointDetectionMeasurementDocumentItems
    | OpticalImagingMeasurementDocumentItems
)


@dataclass(frozen=True)
class ImageFeature:
    identifier: str
    feature: str
    result: float | InvalidJsonFloat


@dataclass(frozen=True)
class ProcessedData:
    identifier: str
    features: list[ImageFeature]


@dataclass(frozen=True)
class DataSource:
    identifier: str
    feature: str


@dataclass(frozen=True)
class CalculatedDataItem:
    identifier: str
    name: str
    value: float
    unit: str
    data_sources: list[DataSource]


@dataclass(frozen=True)
class Measurement:
    # Measurement metadata
    type_: MeasurementType
    identifier: str
    sample_identifier: str
    location_identifier: str
    analyst: str | None = None
    measurement_time: str | None = None
    well_plate_identifier: str | None = None

    # Measurements
    absorbance: JsonFloat | None = None
    fluorescence: JsonFloat | None = None
    luminescence: JsonFloat | None = None

    # Processed data
    processed_data: ProcessedData | None = None

    # Settings
    detector_wavelength_setting: JsonFloat | None = None
    detector_bandwidth_setting: JsonFloat | None = None
    excitation_wavelength_setting: JsonFloat | None = None
    excitation_bandwidth_setting: JsonFloat | None = None
    wavelength_filter_cutoff_setting: float | None = None
    detector_distance_setting: float | None = None
    scan_position_setting: ScanPositionSettingPlateReader | None = None
    detector_gain_setting: str | None = None
    detector_carriage_speed: str | None = None
    compartment_temperature: float | None = None
    number_of_averages: float | None = None

    # Optical imaging
    exposure_duration_setting: float | None = None
    illumination_setting: float | None = None
    magnification_setting: float | None = None
    transmitted_light_setting: TransmittedLightSetting | None = None
    auto_focus_setting: bool | None = None
    image_count_setting: float | None = None
    fluorescent_tag_setting: str | None = None


@dataclass(frozen=True)
class MeasurementGroup:
    measurements: list[Measurement]
    plate_well_count: float
    analytical_method_identifier: str | None = None
    experimental_data_identifier: str | None = None
    _measurement_time: str | None = None
    processed_data: ProcessedData | None = None

    @property
    def measurement_time(self) -> str | None:
        if self._measurement_time is not None:
            return self._measurement_time
        if (
            self.measurements
            and len({m.measurement_time for m in self.measurements}) == 1
            and self.measurements[0].measurement_time
        ):
            return self.measurements[0].measurement_time
        return None


@dataclass(frozen=True)
class Metadata:
    device_identifier: str
    device_type: str
    model_number: str
    software_name: str
    detection_type: str | None = None
    unc_path: str | None = None
    software_version: str | None = None
    equipment_serial_number: str | None = None
    product_manufacturer: str | None = None

    file_name: str | None = None
    data_system_instance_id: str | None = None

    analyst: str | None = None
    measurement_time: str | None = None


@dataclass(frozen=True)
class Data:
    metadata: Metadata
    measurement_groups: list[MeasurementGroup]
    calculated_data: list[CalculatedDataItem] | None = None


class Mapper:
    MANIFEST = "http://purl.allotrope.org/manifests/plate-reader/BENCHLING/2023/09/plate-reader.manifest"

    def __init__(
        self, asm_converter_name: str, get_date_time: Callable[[str], TDateTimeValue]
    ) -> None:
        self.converter_name = asm_converter_name
        self.get_date_time = get_date_time

    def map_model(self, data: Data) -> Model:
        return Model(
            plate_reader_aggregate_document=PlateReaderAggregateDocument(
                device_system_document=DeviceSystemDocument(
                    device_identifier=data.metadata.device_identifier,
                    model_number=data.metadata.model_number,
                    equipment_serial_number=data.metadata.equipment_serial_number,
                    product_manufacturer=data.metadata.product_manufacturer,
                ),
                data_system_document=DataSystemDocument(
                    data_system_instance_identifier=data.metadata.data_system_instance_id,
                    file_name=data.metadata.file_name,
                    UNC_path=data.metadata.unc_path,
                    software_name=data.metadata.software_name,
                    software_version=data.metadata.software_version,
                    ASM_converter_name=self.converter_name,
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                ),
                plate_reader_document=[
                    self._get_technique_document(measurement_group, data.metadata)
                    for measurement_group in data.measurement_groups
                ],
                calculated_data_aggregate_document=self._get_calculated_data_aggregate_document(
                    data.calculated_data
                ),
            ),
            field_asm_manifest=self.MANIFEST,
        )

    def _get_technique_document(
        self, measurement_group: MeasurementGroup, metadata: Metadata
    ) -> PlateReaderDocumentItem:
        return PlateReaderDocumentItem(
            analyst=metadata.analyst,
            measurement_aggregate_document=MeasurementAggregateDocument(
                analytical_method_identifier=measurement_group.analytical_method_identifier,
                experimental_data_identifier=measurement_group.experimental_data_identifier,
                container_type=ContainerType.well_plate,
                plate_well_count=TQuantityValueNumber(
                    value=measurement_group.plate_well_count
                ),
                measurement_time=self.get_date_time(
                    assert_not_none(
                        measurement_group.measurement_time or metadata.measurement_time,
                        msg="Failed to parse valid timestamp",
                    )
                ),
                measurement_document=[
                    self._get_measurement_document(measurement, metadata)
                    for measurement in measurement_group.measurements
                ],
                processed_data_aggregate_document=self._get_processed_data_aggregate_document(
                    measurement_group.processed_data
                ),
            ),
        )

    def _get_measurement_document(
        self, measurement: Measurement, metadata: Metadata
    ) -> MeasurementDocumentItems:
        # TODO(switch-statement): use switch statement once Benchling can use 3.10 syntax
        if measurement.type_ == MeasurementType.OPTICAL_IMAGING:
            return self._get_optical_imaging_measurement_document(measurement, metadata)
        elif measurement.type_ == MeasurementType.ULTRAVIOLET_ABSORBANCE:
            return self._get_ultraviolet_absorbance_measurement_document(
                measurement, metadata
            )
        elif measurement.type_ == MeasurementType.LUMINESCENCE:
            return self._get_luminescence_measurement_document(measurement, metadata)
        elif measurement.type_ == MeasurementType.FLUORESCENCE:
            return self._get_fluorescence_measurement_document(measurement, metadata)
        else:
            msg = f"Unexpected measurement type: {measurement.type}"
            raise AllotropyParserError(msg)

    def _get_optical_imaging_measurement_document(
        self, measurement: Measurement, metadata: Metadata
    ) -> OpticalImagingMeasurementDocumentItems:
        return OpticalImagingMeasurementDocumentItems(
            measurement_identifier=measurement.identifier,
            sample_document=self._get_sample_document(measurement),
            device_control_aggregate_document=OpticalImagingDeviceControlAggregateDocument(
                device_control_document=[
                    OpticalImagingDeviceControlDocumentItem(
                        device_type=metadata.device_type,
                        detection_type=metadata.detection_type,
                        detector_wavelength_setting=quantity_or_none(
                            TQuantityValueNanometer,
                            measurement.detector_wavelength_setting,
                        ),
                        exposure_duration_setting=quantity_or_none(
                            TQuantityValueMilliSecond,
                            measurement.exposure_duration_setting,
                        ),
                        illumination_setting=quantity_or_none(
                            TQuantityValueUnitless, measurement.illumination_setting
                        ),
                        detector_gain_setting=measurement.detector_gain_setting,
                        magnification_setting=quantity_or_none(
                            TQuantityValueUnitless,
                            measurement.magnification_setting,
                        ),
                        transmitted_light_setting=measurement.transmitted_light_setting,
                        auto_focus_setting=measurement.auto_focus_setting,
                        fluorescent_tag_setting=measurement.fluorescent_tag_setting,
                        image_count_setting=quantity_or_none(
                            TQuantityValueUnitless,
                            measurement.image_count_setting,
                        ),
                        excitation_wavelength_setting=quantity_or_none(
                            TQuantityValueNanometer,
                            measurement.excitation_wavelength_setting,
                        ),
                    )
                ]
            ),
            processed_data_aggregate_document=self._get_processed_data_aggregate_document(
                measurement.processed_data
            ),
        )

    def _get_ultraviolet_absorbance_measurement_document(
        self, measurement: Measurement, metadata: Metadata
    ) -> UltravioletAbsorbancePointDetectionMeasurementDocumentItems:
        return UltravioletAbsorbancePointDetectionMeasurementDocumentItems(
            measurement_identifier=measurement.identifier,
            sample_document=self._get_sample_document(measurement),
            device_control_aggregate_document=UltravioletAbsorbancePointDetectionDeviceControlAggregateDocument(
                device_control_document=[
                    UltravioletAbsorbancePointDetectionDeviceControlDocumentItem(
                        device_type=metadata.device_type,
                        detection_type=metadata.detection_type,
                        detector_wavelength_setting=TQuantityValueNanometer(
                            value=assert_not_none(  # type: ignore[arg-type]
                                measurement.detector_wavelength_setting,
                                msg="Missing wavelength setting value in ultraviolet absorbance measurement",
                            )
                        ),
                        number_of_averages=quantity_or_none(
                            TQuantityValueNumber, measurement.number_of_averages
                        ),
                        detector_carriage_speed_setting=measurement.detector_carriage_speed,
                    )
                ]
            ),
            absorbance=TQuantityValueMilliAbsorbanceUnit(
                value=assert_not_none(  # type: ignore[arg-type]
                    value=measurement.absorbance,
                    msg="Missing absorbance value in ultraviolet absorbance measurement",
                )
            ),
            compartment_temperature=quantity_or_none(
                TQuantityValueDegreeCelsius, measurement.compartment_temperature
            ),
        )

    def _get_luminescence_measurement_document(
        self, measurement: Measurement, metadata: Metadata
    ) -> LuminescencePointDetectionMeasurementDocumentItems:
        return LuminescencePointDetectionMeasurementDocumentItems(
            measurement_identifier=measurement.identifier,
            sample_document=self._get_sample_document(measurement),
            device_control_aggregate_document=LuminescencePointDetectionDeviceControlAggregateDocument(
                device_control_document=[
                    LuminescencePointDetectionDeviceControlDocumentItem(
                        device_type=metadata.device_type,
                        detection_type=metadata.detection_type,
                        detector_wavelength_setting=quantity_or_none(
                            TQuantityValueNanometer,
                            measurement.detector_wavelength_setting,
                        ),
                        detector_bandwidth_setting=quantity_or_none(
                            TQuantityValueNanometer,
                            measurement.detector_bandwidth_setting,
                        ),
                        detector_distance_setting__plate_reader_=quantity_or_none(
                            TQuantityValueMillimeter,
                            measurement.detector_distance_setting,
                        ),
                        scan_position_setting__plate_reader_=measurement.scan_position_setting,
                        detector_gain_setting=measurement.detector_gain_setting,
                        detector_carriage_speed_setting=measurement.detector_carriage_speed,
                    )
                ]
            ),
            luminescence=TQuantityValueRelativeLightUnit(
                value=assert_not_none(  # type: ignore[arg-type]
                    measurement.luminescence,
                    msg="Missing luminescence value in luminescence measurement",
                )
            ),
        )

    def _get_fluorescence_measurement_document(
        self, measurement: Measurement, metadata: Metadata
    ) -> FluorescencePointDetectionMeasurementDocumentItems:
        return FluorescencePointDetectionMeasurementDocumentItems(
            measurement_identifier=measurement.identifier,
            device_control_aggregate_document=FluorescencePointDetectionDeviceControlAggregateDocument(
                device_control_document=[
                    FluorescencePointDetectionDeviceControlDocumentItem(
                        device_type=metadata.device_type,
                        detection_type=metadata.detection_type,
                        detector_wavelength_setting=quantity_or_none(
                            TQuantityValueNanometer,
                            measurement.detector_wavelength_setting,
                        ),
                        detector_bandwidth_setting=quantity_or_none(
                            TQuantityValueNanometer,
                            measurement.detector_bandwidth_setting,
                        ),
                        excitation_wavelength_setting=quantity_or_none(
                            TQuantityValueNanometer,
                            measurement.excitation_wavelength_setting,
                        ),
                        excitation_bandwidth_setting=quantity_or_none(
                            TQuantityValueNanometer,
                            measurement.excitation_bandwidth_setting,
                        ),
                        wavelength_filter_cutoff_setting=quantity_or_none(
                            TQuantityValueNanometer,
                            measurement.wavelength_filter_cutoff_setting,
                        ),
                        detector_distance_setting__plate_reader_=quantity_or_none(
                            TQuantityValueMillimeter,
                            measurement.detector_distance_setting,
                        ),
                        scan_position_setting__plate_reader_=measurement.scan_position_setting,
                        detector_gain_setting=measurement.detector_gain_setting,
                        number_of_averages=quantity_or_none(
                            TQuantityValueNumber, measurement.number_of_averages
                        ),
                        detector_carriage_speed_setting=measurement.detector_carriage_speed,
                    )
                ]
            ),
            fluorescence=TQuantityValueRelativeFluorescenceUnit(
                value=assert_not_none(  # type: ignore[arg-type]
                    measurement.fluorescence,
                    msg="Missing fluorescence value in fluorescence measurement",
                )
            ),
            compartment_temperature=quantity_or_none(
                TQuantityValueDegreeCelsius, measurement.compartment_temperature
            ),
            sample_document=self._get_sample_document(measurement),
        )

    def _get_sample_document(self, measurement: Measurement) -> SampleDocument:
        return SampleDocument(
            sample_identifier=measurement.sample_identifier,
            location_identifier=measurement.location_identifier,
            well_plate_identifier=measurement.well_plate_identifier,
        )

    def _get_processed_data_aggregate_document(
        self, data: ProcessedData | None
    ) -> ProcessedDataAggregateDocument | None:
        if not data:
            return None

        # NOTE: this / ProcessedData operating only on "image features" is almost certainly not comprehensive,
        # and should be expanded as needed.
        return ProcessedDataAggregateDocument(
            processed_data_document=[
                ProcessedDataDocumentItem(
                    processed_data_identifier=data.identifier,
                    image_feature_aggregate_document=ImageFeatureAggregateDocument(
                        image_feature_document=[
                            ImageFeatureDocumentItem(
                                image_feature_identifier=image_feature.identifier,
                                image_feature_name=image_feature.feature,
                                image_feature_result=TQuantityValueUnitless(
                                    value=image_feature.result
                                ),
                            )
                            for image_feature in data.features
                        ]
                    ),
                )
            ]
        )

    def _get_calculated_data_aggregate_document(
        self, calculated_data_items: list[CalculatedDataItem] | None
    ) -> CalculatedDataAggregateDocument | None:
        if not calculated_data_items:
            return None

        return CalculatedDataAggregateDocument(
            calculated_data_document=[
                CalculatedDataDocumentItem(
                    calculated_data_identifier=calculated_data_item.identifier,
                    calculated_data_name=calculated_data_item.name,
                    calculated_result=TQuantityValue(
                        value=calculated_data_item.value,
                        unit=calculated_data_item.unit,
                    ),
                    data_source_aggregate_document=DataSourceAggregateDocument(
                        data_source_document=[
                            DataSourceDocumentItem(
                                data_source_identifier=item.identifier,
                                data_source_feature=item.feature,
                            )
                            for item in calculated_data_item.data_sources
                        ]
                    ),
                )
                for calculated_data_item in calculated_data_items
            ]
        )



""" Parser file for ThermoFisher Qubit 4 Adapter """

from allotropy.allotrope.models.adm.spectrophotometry.benchling._2023._12.spectrophotometry import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.spectrophotometry.benchling._2023._12.spectrophotometry import (
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.thermo_fisher_qubit4 import constants
from allotropy.parsers.thermo_fisher_qubit4.thermo_fisher_qubit4_reader import (
    ThermoFisherQubit4Reader,
)
from allotropy.parsers.thermo_fisher_qubit4.thermo_fisher_qubit4_structure import (
    create_data,
)
from allotropy.parsers.vendor_parser import VendorParser


class ThermoFisherQubit4Parser(VendorParser):
    """
    Parser for the ThermoFisher Qubit 4 data files.

    This parser reads data from ThermoFisher Qubit 4 files and converts it into an Allotrope model. The main functionalities
    include extracting and converting specific measurement and device control data, as well as handling custom sample and
    device information.
    """

    @property
    def display_name(self) -> str:
        return constants.DISPLAY_NAME

    @property
    def release_state(self) -> ReleaseState:
        return ReleaseState.RECOMMENDED

    def to_allotrope(self, named_file_contents: NamedFileContents) -> Model:
        df = ThermoFisherQubit4Reader.read(named_file_contents)
        data = create_data(df, named_file_contents.original_file_name)
        return self._get_mapper(Mapper).map_model(data)



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from io import StringIO
import re

import pandas as pd

from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueNumber,
)
from allotropy.allotrope.models.shared.definitions.definitions import InvalidJsonFloat
from allotropy.allotrope.pandas_util import read_csv
from allotropy.parsers.lines_reader import CsvReader
from allotropy.parsers.utils.pandas import SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import assert_not_none


class ReadType(Enum):
    ABSORBANCE = "Absorbance"
    FLUORESCENCE = "Fluorescence"


@dataclass(frozen=True)
class Header:
    user: str
    test_name: str
    date: str
    time: str
    id1: str
    id2: str | None
    id3: str | None
    path: str | None
    test_id: str | None

    @staticmethod
    def create(reader: list[str]) -> Header:
        csv_stream = StringIO("\n".join(reader))
        raw_data = read_csv(csv_stream, header=None)
        df = pd.melt(raw_data, value_vars=raw_data.columns.to_list()).dropna(
            axis="index"
        )
        new = df["value"].str.split(": ", expand=True, n=1)
        data = SeriesData(pd.Series(new[1].values, index=new[0].str.upper()))
        return Header(
            user=assert_not_none(data.get(str, "USER"), msg="User not found in file."),
            path=data.get(str, "PATH"),
            test_id=data.get(str, "TEST ID"),
            test_name=assert_not_none(
                data.get(str, "TEST NAME"), msg="User not found in file."
            ),
            date=assert_not_none(
                data.get(str, "DATE"), msg="Datestamp not found in file."
            ),
            time=assert_not_none(
                data.get(str, "TIME"), msg="Timestamp not found in file."
            ),
            id1=assert_not_none(data.get(str, "ID1"), msg="ID1 not found in file."),
            id2=data.get(str, "ID2"),
            id3=data.get(str, "ID3"),
        )


@dataclass(frozen=True)
class Wavelength:
    wavelength: float
    ex_wavelength: float | InvalidJsonFloat

    @staticmethod
    def create(csv_data: list[str]) -> Wavelength:
        raw_wavelengths = assert_not_none(
            re.search(
                r"Raw Data \((?P<wavelength1>\d+)(?:/)?(?P<wavelength2>\d+)?(?:\))",
                "\n".join(csv_data),
            ),
            msg="Wavelengths not found in input file.",
        )
        if raw_wavelengths.group("wavelength2"):
            return Wavelength(
                wavelength=float(raw_wavelengths.group("wavelength2")),
                ex_wavelength=float(raw_wavelengths.group("wavelength1")),
            )
        else:  # wavelength 1 only
            return Wavelength(
                wavelength=float(raw_wavelengths.group("wavelength1")),
                ex_wavelength=InvalidJsonFloat.NaN,
            )


def get_plate_data(csv_data: list[str]) -> pd.DataFrame:
    csv_reader = CsvReader(csv_data)
    raw_data = assert_not_none(
        csv_reader.lines_as_df(csv_data, skiprows=2),
        msg="Dataframe not found.",
    )
    raw_data.rename(columns={0: "row"}, inplace=True)
    data = raw_data.melt(id_vars=["row"], var_name="col", value_name="value")
    data.dropna(inplace=True)
    data["uuid"] = [random_uuid_str() for _ in range(len(data))]
    return data


def get_plate_well_count(csv_data: list[str]) -> TQuantityValueNumber:
    if re.search(r"23,24\nA", "\n".join(csv_data)):
        plate_well_count = 384
    if re.search(r"11,12\nA", "\n".join(csv_data)):
        plate_well_count = 384
    return TQuantityValueNumber(value=plate_well_count)



from collections import defaultdict
from collections.abc import Callable
import copy
import json
from pathlib import Path
import re
from typing import Any
import urllib.parse

from allotropy.allotrope.schema_parser.update_units import unit_name_from_iri
from allotropy.allotrope.schemas import (
    get_schema,
    get_shared_definitions,
    get_shared_unit_definitions,
)


def _is_array_schema(schema: dict[str, Any]) -> bool:
    return isinstance(schema, dict) and "items" in schema


def _is_direct_object_schema(schema: dict[str, Any]) -> bool:
    return (
        isinstance(schema, dict)
        and "type" in schema
        and schema["type"] == "object"
        or any(key in schema for key in ["properties", "required"])
    )


def _is_composed_object_schema(schema: dict[str, Any]) -> bool:
    return isinstance(schema, dict) and any(
        key in schema for key in ["anyOf", "oneOf", "allOf"]
    )


def _is_object_schema(schema: dict[str, Any]) -> bool:
    return _is_direct_object_schema(schema) or _is_composed_object_schema(schema)


def _create_object_schema(
    properties: dict[str, Any], required: list[str]
) -> dict[str, Any]:
    schema = {"properties": properties}
    if required:
        schema["required"] = required  # type: ignore[assignment]
    return schema


def _is_class_schema(schema: dict[str, Any]) -> bool:
    return _is_object_schema(schema) or _is_array_schema(schema)


def _is_ref_schema(schema: dict[str, Any]) -> bool:
    return isinstance(schema, dict) and "$ref" in schema


def _is_ref_schema_array(schema: dict[str, Any]) -> bool:
    return _is_array_schema(schema) and _is_ref_schema(schema["items"])


def _escape_def_name(def_name: str) -> str:
    return urllib.parse.unquote(def_name).replace("~1", "/")


def _get_def_name(reference: str) -> str:
    return _escape_def_name(reference.split("/")[-1])


def _get_reference_from_url(value: Any) -> tuple[str | None, str | None]:
    # Identify URL-like references, and return a sanitized version that can be followed by generation script.
    # Return schema_name and definition name separately for use in other logic.
    ref_match = re.match(
        r"http://purl.allotrope.org/json-schemas/(.*schema)(\#/\$defs/)?(.*)?",
        str(value),
    )
    if not ref_match:
        return None, None
    schema_name, _, def_name = ref_match.groups()
    schema_name = re.subn(r"[\/\-\.]", "_", schema_name)[0]
    return schema_name, _escape_def_name(def_name)


def _get_required(schema: dict[str, Any]) -> list[str]:
    if _is_array_schema(schema):
        required = schema["items"].get("required", [])
    else:
        required = schema.get("required", [])
    if not isinstance(required, list):
        msg = f"Invalid items in schema: {schema}"
        raise AssertionError(msg)
    return required


def _all_values_equal(values: list[Any]) -> bool:
    return len(values) <= 1 or all(value == values[0] for value in values[1:])


def _should_filter_key(key: str) -> bool:
    return key in (
        "if",
        "then",
        "prefixItems",
        "contains",
        "$comment",
        "minItems",
        "maxItems",
    )


def _should_skip_key(key: str) -> bool:
    return key in ("$defs", "$custom")


def _validate_dictionary_type(dict_value: Any) -> dict[str, Any]:
    if not isinstance(dict_value, dict):
        msg = f"Invalid result, expected dictionary: {dict_value}"
        raise AssertionError(msg)
    return dict_value


def _powerset_indices_from_index(index: int) -> set[int]:
    # Returns the list of indices that appear in the permutation of the powerset at index.
    # Does this by doing the binary expansion of the index, and then return the digits that are 1.
    return {j for j, digit in enumerate(f"{index:b}"[::-1]) if digit == "1"}


class SchemaCleaner:
    def __init__(self) -> None:
        # These are used to track unit references for auto-generating units file.
        self.unit_to_name: dict[str, str] = {}
        self.unit_to_iri: dict[str, str] = {}
        self.referenced_units: set[str] = set()
        for unit_schema in get_shared_unit_definitions().values():
            self._add_unit(
                unit=unit_schema["properties"]["unit"]["const"],
                unit_iri=unit_schema["properties"]["unit"]["$asm.unit-iri"],
            )
        # These are used to track definitions from shared/..., to properly replace the references.
        self.replaced_definitions: defaultdict[str, list[str]] = defaultdict(list)
        self.definitions = get_shared_definitions()

        # These are only used when cleaning definition schemas, and are needed to properly
        # dereference definition references. Though this smells, the flag is nested so deep in the
        # code, that it would be impractical to send this all the way down, and I think this is the
        # lesser evil.
        self.enclosing_schema_name: str | None = None
        self.enclosing_schema_keys: dict[str, Any] | None = None

    def _is_unit_name_ref(self, ref: str) -> bool:
        return _get_def_name(ref) in self.unit_to_name.values()

    def _add_unit(self, unit: str, unit_iri: str) -> None:
        self.unit_to_name[unit] = unit_name_from_iri(unit_iri)
        self.unit_to_iri[unit] = unit_iri

    def _is_quantity_value(self, values: list[dict[str, Any]]) -> bool:
        # Check if this schema is a special case of allOf: [tQuantityValue, unit] and if so replace.
        if len(values) != 2:  # noqa: PLR2004
            return False
        quantity_value_found = False
        unit_found = False
        for value in values:
            if not _is_ref_schema(value):
                return False
            if _get_def_name(value["$ref"]) == "tQuantityValue":
                quantity_value_found = True
            if self._is_unit_name_ref(value["$ref"]):
                unit_found = True
        return quantity_value_found and unit_found

    def _fix_quantity_value_reference(
        self, values: list[dict[str, Any]]
    ) -> dict[str, Any]:
        for value in values:
            def_name = _get_def_name(value["$ref"])
            if def_name == "tQuantityValue":
                continue
            return {"$ref": f"#/$defs/tQuantityValue{def_name}"}

        msg = f"Failed to find value unit in quantity value reference: {values}"
        raise AssertionError(msg)

    def _try_combine_object_schemas(
        self, schemas: list[dict[str, Any]]
    ) -> dict[str, Any]:
        # Try to combine a flattened list of object schemas
        if any(_is_array_schema(schema) for schema in schemas):
            msg = "Unexpected array schema in _try_combine_object_schemas"
            raise AssertionError(msg)

        if any(_is_composed_object_schema(schema) for schema in schemas):
            msg = "Unexpected composed object schema in _try_combine_object_schemas"
            raise AssertionError(msg)

        all_values = defaultdict(list)
        for schema in schemas:
            for key, value in schema.get("properties", {}).items():
                all_values[key].append(value)

        combined_props = {}
        for key, values in dict(all_values).items():
            if _all_values_equal(values):
                combined_props[key] = values[0]
            elif all(_is_class_schema(value) for value in values):
                combined_props[key] = self._combine_allof(values)
            elif any(
                _is_ref_schema(value) or _is_ref_schema_array(value) for value in values
            ):
                combined_props[key] = self._combine_allof(
                    self._dereference_values(values)
                )
            else:
                msg = f"Error combining schemas, conflicting values for key '{key}': {[f'{value}' for value in values]}"
                raise AssertionError(msg)

        return _create_object_schema(
            combined_props,
            sorted(set.union(*[set(schema.get("required", [])) for schema in schemas])),
        )

    def _try_combine_schemas(self, schemas: list[dict[str, Any]]) -> dict[str, Any]:
        schemas = self._flatten_schemas(schemas)

        # When combining schemas, we need to detect anyOf as we do when cleaning allOf, but here we force
        # combining values afterwards.
        if any("anyOf" in schema for schema in schemas):
            schemas = self._invert_allof(schemas, "anyOf")["anyOf"]
            return {
                "anyOf": [self._combine_allof(schema["allOf"]) for schema in schemas]
            }

        # Combine array schemas by combining their inner values.
        if any(_is_array_schema(schema) for schema in schemas):
            if not all(_is_array_schema(schema) for schema in schemas):
                msg = f"Could not combine array and object schemas: {schemas}"
                raise AssertionError(msg)
            return {
                "items": self._try_combine_schemas(
                    [schema["items"] for schema in schemas]
                )
            }

        return self._try_combine_object_schemas(schemas)

    def _flatten_schema(self, value: dict[str, Any]) -> dict[str, Any]:
        # Flattens a composed schema into a single list of schemas and combines them.
        if not _is_composed_object_schema(value):
            return value

        value = copy.deepcopy(value)
        allof_values = value.pop("allOf", [])
        if "anyOf" in value:
            allof_values.append({"anyOf": value.pop("anyOf")})
        if "oneOf" in value:
            if len(value["oneOf"]) > 1:
                msg = "Cannot flatten oneOf with more than one value in _flatten_schema"
                raise AssertionError(msg)
            allof_values.append(value.pop("oneOf")[0])

        if _is_class_schema(value):
            allof_values.append(value)

        if len(allof_values) == 1 and "allOf" not in allof_values[0]:
            return _validate_dictionary_type(allof_values[0])

        return _validate_dictionary_type(self._combine_allof_schemas(allof_values))

    def _flatten_schemas(self, values: list[dict[str, Any]]) -> list[dict[str, Any]]:
        return [self._flatten_schema(value) for value in values]

    def _dereference_value(self, value: dict[str, Any]) -> dict[str, Any]:
        # From a reference schema ({"$ref": <pointer>}) returns the specified schema
        result = copy.deepcopy(value)
        if _is_ref_schema(value):
            if (
                self._is_unit_name_ref(value["$ref"])
                or "tQuantityValue" in value["$ref"]
            ):
                return result
            # NOTE: assumes ref is cleaned.
            # NOTE: we do not further dereference values because we don't want to remove all
            # definitions and mess up class inheritance. We will do so if needed in combine_schemas
            reference = result.pop("$ref")
            match = re.match(r"\#/\$defs/(\w*)(?:/\$defs/)?(\w*)?", reference)
            if not match:
                msg = f"Invalid reference, it may not have been cleaned: {reference}"
                raise AssertionError(msg)
            def_1, def_2 = match.groups()
            if def_2:
                result |= self._dereference_value(
                    self.definitions[def_1]["$defs"][def_2]
                )
            else:
                result |= self._dereference_value(self.definitions[def_1])
        for group_key in ["anyOf", "allOf", "oneOf"]:
            if group_key in value:
                result[group_key] = self._dereference_values(result[group_key])
        if "properties" in value:
            result["properties"] = self._dereference_value(value["properties"])
        if "items" in value:
            result["items"] = self._dereference_value(value["items"])
        return result

    def _dereference_values(self, values: list[dict[str, Any]]) -> list[dict[str, Any]]:
        return [self._dereference_value(value) for value in values]

    def _required_anywhere(self, value: Any) -> bool:
        # Returns True is "required" key is found anywhere in a value recursively.
        # Used to check if there are any reqiured keys in a schema to determine if it should be combined.
        if _is_ref_schema(value):
            if (
                self._is_unit_name_ref(value["$ref"])
                or "QuantityValue" in value["$ref"]
            ):
                return False
            return self._required_anywhere(self._dereference_value(value))
        if isinstance(value, dict):
            return any(
                k == "required" or self._required_anywhere(v) for k, v in value.items()
            )
        elif isinstance(value, list):
            return any(self._required_anywhere(v) for v in value)
        return False

    def _combine_anyof_schemas(
        self, schemas: list[dict[str, Any]]
    ) -> list[dict[str, Any]]:
        # For schemas without required values, we want to combine them as much as possible, but sometimes
        # can't due to keys with conflcting values. To do this, we take the power set of combinations and
        # try to combine each. When we succeed, remove other sets that are covered by that powerset.
        any_required_keys = self._required_anywhere(schemas)
        if not self._required_anywhere(schemas):
            try:
                return [self._try_combine_schemas(schemas)]
            except AssertionError:
                pass

        successful: list[tuple[dict[str, Any], set[int]]] = []
        # total = 2**len(schemas) - 1
        for index in range(2 ** len(schemas) - 1, 0, -1):
            indices = _powerset_indices_from_index(index)
            # If there are not any required keys we can check if the set is covered first.
            if not any_required_keys and any(
                indices.issubset(schema_indices) for _, schema_indices in successful
            ):
                continue

            # If there are required keys, we need to combine the schemas first so we can cross check the required
            # keys against covering sets.
            try:
                combined = self._try_combine_schemas([schemas[j] for j in indices])
            except AssertionError:
                continue
            covered = False
            for schema, schema_indices in successful:
                if indices.issubset(schema_indices) and _get_required(
                    combined
                ) == _get_required(schema):
                    covered = True
                    break
            if not covered:
                successful.append((combined, indices))

        return [schema for schema, _ in successful]

    def _combine_anyof(self, values: list[Any]) -> dict[str, Any]:
        if all(_is_class_schema(value) for value in values):
            values = self._combine_anyof_schemas(values)

        return _validate_dictionary_type(
            {"anyOf": values} if len(values) > 1 else values[0]
        )

    def _invert_allof(self, schemas: list[dict[str, Any]], key: str) -> dict[str, Any]:
        new_schemas: list[list[dict[str, Any]]] = []
        for schema in schemas:
            schemas_list = schema[key] if key in schema else [schema]
            new_schemas = (
                [
                    [*new_schema, add_schema]
                    for new_schema in new_schemas
                    for add_schema in schemas_list
                ]
                if new_schemas
                else [[schema] for schema in schemas_list]
            )

        return {key: [{"allOf": schemas} for schemas in new_schemas]}

    def _combine_allof_schemas(
        self, schemas: list[dict[str, Any]]
    ) -> dict[str, Any] | list[dict[str, Any]]:
        schemas = self._clean_value(schemas)
        if not all(_is_class_schema(schema) for schema in schemas):
            if any(_is_class_schema(schema) for schema in schemas):
                msg = f"_combine_allof_schemas can only be called with a list of object schema dictionaries: {schemas}"
                raise AssertionError(msg)

            return schemas

        return self._try_combine_schemas(schemas)

    def _combine_allof(self, values: list[Any]) -> dict[str, Any]:
        if not all(isinstance(value, dict) for value in values):
            msg = "Unhandled case: expected every item in an allOf to be a dictionary"
            raise AssertionError(msg)

        if _all_values_equal(values):
            return _validate_dictionary_type(values[0])

        # If we are trying to combine tQuantityValue with tQuantityValue units, we can do so if there are
        # not conflicting units.
        if all(
            _is_ref_schema(schema) and "QuantityValue" in schema["$ref"]
            for schema in values
        ):
            unique = {_get_def_name(schema["$ref"]) for schema in values} - {
                "tQuantityValue"
            }
            if len(unique) == 1:
                return {"$ref": f"#/$defs/{next(iter(unique))}"}
            msg = f"Unable to combine multiple different tQuantityValue references: {values}"
            raise AssertionError(msg)

        # datamodel-codegen can not handle oneOf nested inside allOf. Fix this by reversing the order,
        # making a oneOf with each possible product of allOf
        if any("oneOf" in value for value in values):
            return self._clean_schema(self._invert_allof(values, "oneOf"))

        # This must come after fixing oneOf because oneOf may break into multiple quantity values.
        if self._is_quantity_value(values):
            return self._fix_quantity_value_reference(values)

        # Deference values and check for oneOf/anyOf inversion again.
        derefed_values = self._dereference_values(values)
        if any("oneOf" in value for value in derefed_values):
            return self._clean_schema(self._invert_allof(derefed_values, "oneOf"))
        if any("anyOf" in value for value in derefed_values):
            return self._clean_schema(self._invert_allof(derefed_values, "anyOf"))

        # If any object in the allOf has required fields, we must combine them in order for the generation
        # script to generate valid dataclasses. This is because dataclass inheritance with optional fields
        # is broken in python < 3.10.
        # If there is an allOf nested with this allOf, combine it.
        if self._required_anywhere(derefed_values) or any(
            "allOf" in schema for schema in derefed_values
        ):
            return _validate_dictionary_type(
                self._combine_allof_schemas(derefed_values)
            )

        # Otherwise, we try to combine the schemas in order to error if there are any conflicting keys,
        # but we don't save the result.
        self._try_combine_schemas(derefed_values)

        return {"allOf": values}

    def _clean_ref_value(self, value: str) -> str:
        # Get the schema and the definition name from the URL to create the local def path.
        schema_name, def_name = _get_reference_from_url(value)

        # If covered by a definition in shared definitions or if a unit, use those.
        if def_name in self.replaced_definitions.get(schema_name or "", []):
            return f"#/$defs/{def_name}"
        elif def_name and def_name in self.unit_to_name:
            self.referenced_units.add(def_name)
            return f"#/$defs/{self.unit_to_name[def_name]}"

        # Otherwise create the local definition path.
        if schema_name and def_name:
            cleaned_ref = f"#/$defs/{schema_name}/$defs/{def_name}"
        elif schema_name or def_name:
            cleaned_ref = f"#/$defs/{schema_name or def_name}"
        else:
            cleaned_ref = value

        # Finally we need to check if this is a path inside another definition schema, and if so use the
        # absolute path.
        def_name = _get_def_name(cleaned_ref)
        if (
            self.enclosing_schema_name
            and self.enclosing_schema_keys
            and def_name not in self.definitions
            and def_name in self.enclosing_schema_keys
        ):
            return f"#/$defs/{self.enclosing_schema_name}/$defs/{def_name}"

        return cleaned_ref

    def _clean_value(
        self,
        value: Any,
        cleaning_function: Callable[[dict[str, Any]], dict[str, Any]] | None = None,
    ) -> Any:
        # Fan out function for handling dict/list values. Calls cleaning_function on dictionaries
        # (assumed to be schemas). Allows passing custom cleaning_fuction to handle cleaning definitions.
        cleaning_function = cleaning_function or self._clean_schema
        if isinstance(value, dict):
            return cleaning_function(value)
        elif isinstance(value, list):
            return list(
                filter(
                    lambda v: bool(v),
                    (self._clean_value(v, cleaning_function) for v in value),
                )
            )
        return value

    def _clean_schema(self, schema: dict[str, Any]) -> dict[str, Any]:
        # If a schema is has properties on it and anyOf/oneOf/allOf composed components, it is essentially
        # an allOf with the parent schema and the rest, combine this way.
        schema = copy.deepcopy(schema)
        if _is_direct_object_schema(schema) and _is_composed_object_schema(schema):
            allof_values = [
                _create_object_schema(
                    schema.pop("properties", {}), schema.pop("required", [])
                ),
                *schema.pop("allOf", []),
            ]
            if "anyOf" in schema:
                allof_values.append({"anyOf": schema.pop("anyOf")})
            if "oneOf" in schema:
                allof_values.append({"oneOf": schema.pop("oneOf")})
            schema["allOf"] = allof_values

        cleaned = {}
        for key, value in schema.items():
            if _should_filter_key(key):
                continue
            if _should_skip_key(key):
                cleaned[key] = value
                continue

            if key == "allOf":
                clean_value = self._clean_value(value)
                cleaned |= self._combine_allof(clean_value)
            elif key == "$ref":
                cleaned[key] = self._clean_ref_value(value)
            else:
                cleaned[key] = self._clean_value(value)

        return {key: value for key, value in cleaned.items() if value not in (None, {})}

    def _clean_def_schema(self, schema: dict[str, Any]) -> dict[str, Any]:
        cleaned = {}
        for key, value in schema.items():
            if _should_filter_key(key):
                continue
            if _should_skip_key(key):
                cleaned[key] = value
                continue

            if key == "$ref":
                cleaned[key] = self._clean_ref_value(value)
            else:
                cleaned[key] = self._clean_value(value, self._clean_def_schema)

        return {key: value for key, value in cleaned.items() if value not in (None, {})}

    def _clean_defs(self, schema: dict[str, Any]) -> dict[str, Any]:
        for schema_name, defs_schema in schema.items():
            # Replace web address reference name with a version that can be followed in datamodel-codegen
            cleaned_schema_name = _get_reference_from_url(schema_name)[0] or schema_name
            # Units are treated specially. We rename the unit (to avoid non-variable names) and store
            # them in separate shared unit schema file, in order to allow for shared imports.
            if cleaned_schema_name.endswith("units_schema"):
                for unit, unit_schema in defs_schema["$defs"].items():
                    self._add_unit(
                        unit=unit,
                        unit_iri=unit_schema["properties"]["unit"]["$asm.unit-iri"],
                    )
            elif "$defs" in defs_schema:
                # Store defs that are replaced with definitions in common/definitions before cleaning.
                # NOTE: this does not attempt to handle futher nested $defs in schemas, but we have not
                # observed that in any schema files yet.
                self.definitions[cleaned_schema_name] = {"$defs": {}}
                for def_name, def_schema in defs_schema["$defs"].items():
                    if self.definitions.get(def_name) == def_schema:
                        self.replaced_definitions[cleaned_schema_name].append(def_name)
                    else:
                        self.definitions[cleaned_schema_name]["$defs"][
                            def_name
                        ] = def_schema
            else:
                self.definitions[cleaned_schema_name] = defs_schema

        cleaned = {}
        for schema_name, defs_schema in schema.items():
            cleaned_schema_name = _get_reference_from_url(schema_name)[0] or schema_name
            if cleaned_schema_name.endswith("units_schema"):
                continue
            elif "$defs" in defs_schema:
                # For references nested inside definitions, we need to change the ref into the full reference
                # path in order for the generation script to find them. e.g.
                #   {"someSchema": {"$ref": "#/$defs/someThing"}}
                #       becomes
                #   {"someSchema": {"$ref": "#/$defs/someSchema/$defs/someThing"}}
                # To accomplish this, we set enclosing_schema_name/keys while cleaning the def schema,
                # and use them in clean_ref_value
                self.enclosing_schema_name = cleaned_schema_name
                self.enclosing_schema_keys = defs_schema["$defs"].keys()
                cleaned_defs = {}
                for def_name, def_schema in defs_schema["$defs"].items():
                    if def_name not in self.replaced_definitions[cleaned_schema_name]:
                        cleaned_defs[def_name] = self._clean_def_schema(def_schema)
                self.enclosing_schema_name = None
                self.enclosing_schema_keys = None
                cleaned[cleaned_schema_name] = {"$defs": cleaned_defs}
            else:
                cleaned[cleaned_schema_name] = self._clean_def_schema(defs_schema)

        self.definitions |= cleaned
        return cleaned

    def get_referenced_units(self) -> dict[str, str]:
        # Returns all units seen in the schema, for auto-generating unit schemas externally.
        return {
            unit: iri
            for unit, iri in self.unit_to_iri.items()
            if unit in self.referenced_units
        }

    def clean(self, schema: dict[str, Any]) -> dict[str, Any]:
        # Call clean defs first, because we store some metadata about overridden definitions that is used in
        # the main body.
        cleaned = copy.deepcopy(schema)

        # Definitions are cleaned differently, because we don't want to change modify them more than we need
        # to. That is, we leave combining compound schemas to the main schema cleaning so we can tell if it
        # is necessary or not. Here, we simply clean up http references and store definitions for later reference.
        cleaned["$defs"] = self._clean_defs(cleaned.get("$defs", {}))

        return self._clean_schema(cleaned)

    def clean_file(self, schema_path: str) -> None:
        schema = self.clean(get_schema(Path(schema_path)))

        with open(schema_path, "w") as f:
            json.dump(schema, f, indent=2)



from collections import defaultdict
from datetime import tzinfo
from enum import Enum
from pathlib import Path

from allotropy.allotrope.schema_parser.path_util import ROOT_DIR
from allotropy.parsers.agilent_gen5.agilent_gen5_parser import AgilentGen5Parser
from allotropy.parsers.agilent_gen5_image.agilent_gen5_image_parser import (
    AgilentGen5ImageParser,
)
from allotropy.parsers.agilent_tapestation_analysis.agilent_tapestation_analysis_parser import (
    AgilentTapestationAnalysisParser,
)
from allotropy.parsers.appbio_absolute_q.appbio_absolute_q_parser import (
    AppbioAbsoluteQParser,
)
from allotropy.parsers.appbio_quantstudio.appbio_quantstudio_parser import (
    AppBioQuantStudioParser,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.appbio_quantstudio_designandanalysis_parser import (
    AppBioQuantStudioDesignandanalysisParser,
)
from allotropy.parsers.beckman_pharmspec.beckman_pharmspec_parser import PharmSpecParser
from allotropy.parsers.beckman_vi_cell_blu.vi_cell_blu_parser import ViCellBluParser
from allotropy.parsers.beckman_vi_cell_xr.vi_cell_xr_parser import ViCellXRParser
from allotropy.parsers.biorad_bioplex_manager.biorad_bioplex_manager_parser import (
    BioradBioplexParser,
)
from allotropy.parsers.bmg_mars.bmg_mars_parser import BmgMarsParser
from allotropy.parsers.chemometec_nucleoview.nucleoview_parser import (
    ChemometecNucleoviewParser,
)
from allotropy.parsers.ctl_immunospot.ctl_immunospot_parser import CtlImmunospotParser
from allotropy.parsers.example_weyland_yutani.example_weyland_yutani_parser import (
    ExampleWeylandYutaniParser,
)
from allotropy.parsers.luminex_xponent.luminex_xponent_parser import (
    LuminexXponentParser,
)
from allotropy.parsers.mabtech_apex.mabtech_apex_parser import MabtechApexParser
from allotropy.parsers.methodical_mind.methodical_mind_parser import (
    MethodicalMindParser,
)
from allotropy.parsers.moldev_softmax_pro.softmax_pro_parser import SoftmaxproParser
from allotropy.parsers.novabio_flex2.novabio_flex2_parser import NovaBioFlexParser
from allotropy.parsers.perkin_elmer_envision.perkin_elmer_envision_parser import (
    PerkinElmerEnvisionParser,
)
from allotropy.parsers.qiacuity_dpcr.qiacuity_dpcr_parser import QiacuitydPCRParser
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.revvity_kaleido.kaleido_parser import KaleidoParser
from allotropy.parsers.roche_cedex_bioht.roche_cedex_bioht_parser import (
    RocheCedexBiohtParser,
)
from allotropy.parsers.roche_cedex_hires.roche_cedex_hires_parser import (
    RocheCedexHiResParser,
)
from allotropy.parsers.thermo_fisher_nanodrop_eight.nanodrop_eight_parser import (
    NanodropEightParser,
)
from allotropy.parsers.thermo_fisher_qubit4.thermo_fisher_qubit4_parser import (
    ThermoFisherQubit4Parser,
)
from allotropy.parsers.thermo_fisher_qubit_flex.thermo_fisher_qubit_flex_parser import (
    ThermoFisherQubitFlexParser,
)
from allotropy.parsers.unchained_labs_lunatic.unchained_labs_lunatic_parser import (
    UnchainedLabsLunaticParser,
)
from allotropy.parsers.utils.timestamp_parser import TimestampParser
from allotropy.parsers.vendor_parser import VendorParser


class Vendor(Enum):
    AGILENT_GEN5 = "AGILENT_GEN5"
    AGILENT_GEN5_IMAGE = "AGILENT_GEN5_IMAGE"
    AGILENT_TAPESTATION_ANALYSIS = "AGILENT_TAPESTATION_ANALYSIS"
    APPBIO_ABSOLUTE_Q = "APPBIO_ABSOLUTE_Q"
    APPBIO_QUANTSTUDIO = "APPBIO_QUANTSTUDIO"
    APPBIO_QUANTSTUDIO_DESIGNANDANALYSIS = "APPBIO_QUANTSTUDIO_DESIGNANDANALYSIS"
    BMG_MARS = "BMG_MARS"
    BECKMAN_PHARMSPEC = "BECKMAN_PHARMSPEC"
    BECKMAN_VI_CELL_BLU = "BECKMAN_VI_CELL_BLU"
    BECKMAN_VI_CELL_XR = "BECKMAN_VI_CELL_XR"
    BIORAD_BIOPLEX = "BIORAD_BIOPLEX"
    CHEMOMETEC_NUCLEOVIEW = "CHEMOMETEC_NUCLEOVIEW"
    CTL_IMMUNOSPOT = "CTL_IMMUNOSPOT"
    EXAMPLE_WEYLAND_YUTANI = "EXAMPLE_WEYLAND_YUTANI"
    LUMINEX_XPONENT = "LUMINEX_XPONENT"
    MABTECH_APEX = "MABTECH_APEX"
    METHODICAL_MIND = "METHODICAL_MIND"
    MOLDEV_SOFTMAX_PRO = "MOLDEV_SOFTMAX_PRO"
    NOVABIO_FLEX2 = "NOVABIO_FLEX2"
    PERKIN_ELMER_ENVISION = "PERKIN_ELMER_ENVISION"
    QIACUITY_DPCR = "QIACUITY_DPCR"
    REVVITY_KALEIDO = "REVVITY_KALEIDO"
    ROCHE_CEDEX_BIOHT = "ROCHE_CEDEX_BIOHT"
    ROCHE_CEDEX_HIRES = "ROCHE_CEDEX_HIRES"
    THERMO_FISHER_NANODROP_EIGHT = "THERMO_FISHER_NANODROP_EIGHT"
    THERMO_FISHER_QUBIT4 = "THERMO_FISHER_QUBIT4"
    THERMO_FISHER_QUBIT_FLEX = "THERMO_FISHER_QUBIT_FLEX"
    UNCHAINED_LABS_LUNATIC = "UNCHAINED_LABS_LUNATIC"

    @property
    def display_name(self) -> str:
        return self.get_parser().display_name

    @property
    def release_state(self) -> ReleaseState:
        return self.get_parser().release_state

    def get_parser(self, default_timezone: tzinfo | None = None) -> VendorParser:
        timestamp_parser = TimestampParser(default_timezone)
        return _VENDOR_TO_PARSER[self](timestamp_parser)


_VENDOR_TO_PARSER: dict[Vendor, type[VendorParser]] = {
    Vendor.AGILENT_GEN5: AgilentGen5Parser,
    Vendor.AGILENT_GEN5_IMAGE: AgilentGen5ImageParser,
    Vendor.AGILENT_TAPESTATION_ANALYSIS: AgilentTapestationAnalysisParser,
    Vendor.APPBIO_ABSOLUTE_Q: AppbioAbsoluteQParser,
    Vendor.APPBIO_QUANTSTUDIO: AppBioQuantStudioParser,
    Vendor.APPBIO_QUANTSTUDIO_DESIGNANDANALYSIS: AppBioQuantStudioDesignandanalysisParser,
    Vendor.BECKMAN_PHARMSPEC: PharmSpecParser,
    Vendor.BECKMAN_VI_CELL_BLU: ViCellBluParser,
    Vendor.BECKMAN_VI_CELL_XR: ViCellXRParser,
    Vendor.BIORAD_BIOPLEX: BioradBioplexParser,
    Vendor.BMG_MARS: BmgMarsParser,
    Vendor.CHEMOMETEC_NUCLEOVIEW: ChemometecNucleoviewParser,
    Vendor.CTL_IMMUNOSPOT: CtlImmunospotParser,
    Vendor.EXAMPLE_WEYLAND_YUTANI: ExampleWeylandYutaniParser,
    Vendor.LUMINEX_XPONENT: LuminexXponentParser,
    Vendor.MABTECH_APEX: MabtechApexParser,
    Vendor.METHODICAL_MIND: MethodicalMindParser,
    Vendor.MOLDEV_SOFTMAX_PRO: SoftmaxproParser,
    Vendor.NOVABIO_FLEX2: NovaBioFlexParser,
    Vendor.PERKIN_ELMER_ENVISION: PerkinElmerEnvisionParser,
    Vendor.QIACUITY_DPCR: QiacuitydPCRParser,
    Vendor.REVVITY_KALEIDO: KaleidoParser,
    Vendor.ROCHE_CEDEX_BIOHT: RocheCedexBiohtParser,
    Vendor.ROCHE_CEDEX_HIRES: RocheCedexHiResParser,
    Vendor.THERMO_FISHER_NANODROP_EIGHT: NanodropEightParser,
    Vendor.THERMO_FISHER_QUBIT4: ThermoFisherQubit4Parser,
    Vendor.THERMO_FISHER_QUBIT_FLEX: ThermoFisherQubitFlexParser,
    Vendor.UNCHAINED_LABS_LUNATIC: UnchainedLabsLunaticParser,
}


def update_readme() -> None:
    release_state_to_parser = defaultdict(set)
    for vendor in Vendor:
        if "example" in str(vendor).lower():
            continue
        release_state_to_parser[vendor.release_state].add(vendor.display_name)

    readme_file = Path(ROOT_DIR, "README.md")
    with open(readme_file) as f:
        contents = f.readlines()

    with open(readme_file, "w") as f:
        in_block = False
        newline_count = 0
        for line in contents:
            if line.startswith("### Recommended"):
                in_block = True
                continue
            if in_block:
                if line == "\n":
                    newline_count += 1
                if newline_count == 3:  # noqa: PLR2004
                    for release_state in [
                        ReleaseState.RECOMMENDED,
                        ReleaseState.CANDIDATE_RELEASE,
                        ReleaseState.WORKING_DRAFT,
                    ]:
                        f.write(
                            f'### {release_state.value.replace("_", " ").title()}\n'
                        )
                        for display_name in sorted(
                            release_state_to_parser.get(release_state, [])
                        ):
                            f.write(f"  - {display_name}\n")
                        f.write("\n")
                    in_block = False
                continue
            f.write(line)



""" Parser file for Roche Cedex HiRes Instrument """
from __future__ import annotations

from decimal import Decimal
from typing import Any, TypeVar

import numpy as np
import pandas as pd

from allotropy.allotrope.converter import add_custom_information_document
from allotropy.allotrope.models.adm.cell_counting.benchling._2023._11.cell_counting import (
    CellCountingAggregateDocument,
    CellCountingDetectorDeviceControlAggregateDocument,
    CellCountingDetectorMeasurementDocumentItem,
    CellCountingDocumentItem,
    DataProcessingDocument,
    DataSystemDocument,
    DeviceControlDocumentItemModel,
    DeviceSystemDocument,
    MeasurementAggregateDocument,
    Model,
    ProcessedDataAggregateDocument,
    ProcessedDataDocumentItem,
    SampleDocument,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueCell,
    TQuantityValueMicroliter,
    TQuantityValueMicrometer,
    TQuantityValueMillionCellsPerMilliliter,
    TQuantityValuePercent,
    TQuantityValueUnitless,
)
from allotropy.constants import ASM_CONVERTER_VERSION
from allotropy.exceptions import AllotropeConversionError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.roche_cedex_hires import constants
from allotropy.parsers.roche_cedex_hires.roche_cedex_hires_reader import (
    RocheCedexHiResReader,
)
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import assert_not_none
from allotropy.parsers.vendor_parser import VendorParser

DataType = TypeVar("DataType")


def get_property_value(
    data_frame: pd.DataFrame, column: str, row: int, datatype: type
) -> DataType | None:
    """
    Retrieves the value from a specified column and row in a DataFrame and converts it
    to the specified datatype.

    Parameters:
    data_frame (pd.DataFrame): The DataFrame from which to retrieve the value.
    column (str): The column name from which to retrieve the value.
    row (int): The row index from which to retrieve the value.
    datatype (type): The type to which the retrieved value should be converted.

    Returns:
    Datatype: The value from the specified cell converted to the specified datatype.
         Returns None if the value is not found.
    """
    return (
        datatype(value=value) if (value := get_value(data_frame, column, row)) else None
    )


def get_value(data_frame: pd.DataFrame, column: str, row: int) -> Any | None:
    """
    Retrieves the value from a specified column and row in a DataFrame, handling NaNs
    and converting certain numpy types to native Python types.

    Parameters:
    data_frame (pd.DataFrame): The DataFrame from which to retrieve the value.
    column (str): The column name from which to retrieve the value.
    row (int): The row index from which to retrieve the value.

    Returns:
    Optional[Any]: The value from the specified cell converted to the appropriate Python type.
                   Returns None if the column does not exist or the value is NaN.
    """
    if column not in data_frame.columns:
        return None
    value = data_frame[column][row]

    if pd.isna(value):
        return None
    if isinstance(value, np.int64):
        return int(value)
    if isinstance(value, np.float64):
        return float(value)
    return value


def get_value_not_none(dataframe: pd.DataFrame, column: str, row: int) -> Any:
    """
    Retrieves the value from a specified column and row in a DataFrame, ensuring the value is not None.

    Parameters:
    dataframe (pd.DataFrame): The DataFrame from which to retrieve the value.
    column (str): The column name from which to retrieve the value.
    row (int): The row index from which to retrieve the value.

    Returns:
    Any: The value from the specified cell.

    Raises:
    AllotropeConversionError: If the value is None.
    """
    value = get_value(dataframe, column, row)
    if value is None:
        msg = f"{constants.VALUE_ERROR} '{column}'."
        raise AllotropeConversionError(msg)
    return value


def get_calculated_data(
    dataframe: pd.DataFrame, column: str, row: int, datatype: Any
) -> Any:
    """
    Retrieves a value from a specified column and row in a DataFrame, converts it to a float
    by dividing it by 1,000,000, and then converts it to the specified datatype.

    Parameters:
    dataframe (pd.DataFrame): The DataFrame from which to retrieve the value.
    column (str): The column name from which to retrieve the value.
    row (int): The row index from which to retrieve the value.
    datatype (Any): The type to which the value should be further converted.

    Returns:
    Any: The value from the specified cell after conversion, or None if the value is not found.
    """
    value = get_value(dataframe, column, row)
    if value is not None:
        value = float(Decimal(str(value)) / Decimal("1000000"))
        return datatype(value=value)
    return value


def get_system_value(data: pd.DataFrame, column_name: str) -> str:
    """
    Retrieves the unique value from a specified column in a DataFrame, ensuring that there
    is only one unique value. If there are multiple unique values, raises an error.

    Parameters:
    data (pd.DataFrame): The DataFrame from which to retrieve the unique value.
    column_name (str): The name of the column from which to retrieve the unique value.

    Returns:
    str: The unique value from the specified column.

    Raises:
    AllotropeConversionError: If there are multiple unique values in the column.
    """
    unique_values = set(data[column_name])
    if len(unique_values) != 1:
        message = f"{constants.MULTIPLE_SYSTEM_ERROR} '{column_name}'"
        raise AllotropeConversionError(message)
    return str(next(iter(unique_values)))


class RocheCedexHiResParser(VendorParser):
    """
    A parser for Roche Cedex HiRes data files, converting them to Allotrope models.

    Properties:
    display_name (str): The display name of the parser.
    release_state (ReleaseState): The current release state of the parser.

    Methods:
    to_allotrope(named_file_contents: NamedFileContents) -> Model:
        Converts the named file contents to an Allotrope model.

    _get_model(data: pd.DataFrame, filename: str) -> Model:
        Constructs the Allotrope model from the parsed data.

    _get_device_system_document(data: pd.DataFrame) -> DeviceSystemDocument:
        Constructs the DeviceSystemDocument from the parsed data.

    _get_cell_counting_document(data: pd.DataFrame) -> list[CellCountingDocumentItem]:
        Constructs a list of CellCountingDocumentItems from the parsed data.

    _get_cell_counting_document_item(data: pd.DataFrame, row: int) -> CellCountingDocumentItem:
        Constructs a CellCountingDocumentItem for a specific row of data.
    """

    @property
    def display_name(self) -> str:
        """
        Returns the display name of the parser.

        Returns:
        str: The display name of the parser.
        """
        return constants.DISPLAY_NAME

    @property
    def release_state(self) -> ReleaseState:
        """
        Returns the current release state of the parser.

        Returns:
        ReleaseState: The current release state of the parser.
        """
        return ReleaseState.RECOMMENDED

    def to_allotrope(self, named_file_contents: NamedFileContents) -> Model:
        """
        Converts the named file contents to an Allotrope model.

        Parameters:
        named_file_contents (NamedFileContents): Contains the contents of instrument export file which is converted to ASM model.

        Returns:
        Model: The resulting Allotrope model.
        """
        return self._get_model(
            data=RocheCedexHiResReader.read(named_file_contents),
            filename=named_file_contents.original_file_name,
        )

    def _get_model(self, data: pd.DataFrame, filename: str) -> Model:
        """
        Constructs the Allotrope model from the parsed data.

        Parameters:
        data (pd.DataFrame): The parsed data.
        filename (str): The original file name.

        Returns:
        Model: The constructed Allotrope model.
        """
        return Model(
            field_asm_manifest="http://purl.allotrope.org/manifests/cell-counting/BENCHLING/2023/11/cell-counting.manifest",
            cell_counting_aggregate_document=CellCountingAggregateDocument(
                device_system_document=self._get_device_system_document(data),
                data_system_document=DataSystemDocument(
                    file_name=filename,
                    software_name=constants.CEDEX_SOFTWARE,
                    ASM_converter_name=self.get_asm_converter_name(),
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                ),
                cell_counting_document=self._get_cell_counting_document(data),
            ),
        )

    def _get_device_system_document(self, data: pd.DataFrame) -> DeviceSystemDocument:
        """
        Constructs the DeviceSystemDocument from the parsed data.

        Parameters:
        data (pd.DataFrame): The parsed data.

        Returns:
        DeviceSystemDocument: The constructed DeviceSystemDocument.
        """
        return DeviceSystemDocument(
            model_number=constants.MODEL_NUMBER,
            product_manufacturer=constants.PRODUCT_MANUFACTURER,
            asset_management_identifier=get_system_value(data, "System name"),
            description=get_system_value(data, "System description"),
            device_identifier=str(get_system_value(data, "Cedex ID")),
            brand_name=constants.BRAND_NAME,
        )

    def _get_cell_counting_document(
        self, data: pd.DataFrame
    ) -> list[CellCountingDocumentItem]:
        """
        Constructs a list of CellCountingDocumentItems from the parsed data.

        Parameters:
        data (pd.DataFrame): The parsed data.

        Returns:
        list[CellCountingDocumentItem]: A list of CellCountingDocumentItems.
        """
        return [
            self._get_cell_counting_document_item(data, i)
            for i in range(len(data.index))
        ]

    def _get_cell_counting_document_item(
        self, data: pd.DataFrame, row: int
    ) -> CellCountingDocumentItem:
        """
        Constructs a CellCountingDocumentItem for a specific row of data.

        Parameters:
        data (pd.DataFrame): The parsed data.
        row (int): The row index of the data.

        Returns:
        CellCountingDocumentItem: The constructed CellCountingDocumentItem.
        """
        sample_custom_document = {
            "group identifier": get_value(data, "Data set name", row),
            "sample draw time": self._get_date_time(
                str(get_value(data, "Sample draw Time", row))
            ),
        }
        processed_data_custom_document = {
            "average compactness": get_property_value(
                data, "Avg Compactness", row, TQuantityValueUnitless
            ),
            "average area": get_property_value(
                data, "Avg Area", row, TQuantityValueUnitless
            ),
            "average perimeter": get_property_value(
                data, "Avg Perimeter", row, TQuantityValueMicrometer
            ),
            "average segment area": get_property_value(
                data, "Avg Segm. Area", row, TQuantityValueUnitless
            ),
            "total object count": get_property_value(
                data, "Total Object Count", row, TQuantityValueCell
            ),
            "standard deviation": get_property_value(
                data, "Std Dev.", row, TQuantityValueCell
            ),
            "aggregate rate": get_property_value(
                data, "Aggregate Rate", row, TQuantityValuePercent
            ),
        }
        return CellCountingDocumentItem(
            analyst=get_value(data, "Username", row),
            measurement_aggregate_document=MeasurementAggregateDocument(
                measurement_document=[
                    CellCountingDetectorMeasurementDocumentItem(
                        measurement_time=self._get_date_time(
                            str(get_value_not_none(data, "Date finished", row))
                        ),
                        measurement_identifier=random_uuid_str(),
                        sample_document=add_custom_information_document(
                            SampleDocument(
                                sample_identifier=get_value_not_none(
                                    data, "Sample identifer", row
                                ),
                                batch_identifier=get_value(
                                    data, "Reactor identifer", row
                                ),
                            ),
                            sample_custom_document,
                        ),
                        device_control_aggregate_document=CellCountingDetectorDeviceControlAggregateDocument(
                            device_control_document=[
                                DeviceControlDocumentItemModel(
                                    device_type=constants.DEVICE_TYPE,
                                    detection_type=constants.DETECTION_TYPE,
                                    sample_volume_setting=get_property_value(
                                        data,
                                        "Sample volume",
                                        row,
                                        TQuantityValueMicroliter,
                                    ),
                                )
                            ]
                        ),
                        processed_data_aggregate_document=ProcessedDataAggregateDocument(
                            processed_data_document=[
                                add_custom_information_document(
                                    ProcessedDataDocumentItem(
                                        data_processing_document=DataProcessingDocument(
                                            cell_type_processing_method=get_value(
                                                data, "Cell type name", row
                                            ),
                                            cell_density_dilution_factor=get_property_value(
                                                data,
                                                "Dilution",
                                                row,
                                                TQuantityValueUnitless,
                                            ),
                                        ),
                                        viability__cell_counter_=TQuantityValuePercent(
                                            value=get_value_not_none(
                                                data, "Viability", row
                                            )
                                        ),
                                        viable_cell_density__cell_counter_=assert_not_none(
                                            get_calculated_data(
                                                data,
                                                "Viable Cell Conc.",
                                                row,
                                                TQuantityValueMillionCellsPerMilliliter,
                                            ),
                                            name="Viable Cell Conc.",
                                        ),
                                        total_cell_density__cell_counter_=get_calculated_data(
                                            data,
                                            "Total Cell Conc.",
                                            row,
                                            TQuantityValueMillionCellsPerMilliliter,
                                        ),
                                        dead_cell_density__cell_counter_=get_calculated_data(
                                            data,
                                            "Dead Cell Conc.",
                                            row,
                                            TQuantityValueMillionCellsPerMilliliter,
                                        ),
                                        total_cell_count=get_property_value(
                                            data,
                                            "Total Cell Count",
                                            row,
                                            TQuantityValueCell,
                                        ),
                                        viable_cell_count=get_property_value(
                                            data,
                                            "Viable Cell Count",
                                            row,
                                            TQuantityValueCell,
                                        ),
                                        dead_cell_count=get_property_value(
                                            data,
                                            "Dead Cell Count",
                                            row,
                                            TQuantityValueCell,
                                        ),
                                        average_live_cell_diameter__cell_counter_=get_property_value(
                                            data,
                                            "Avg Diameter",
                                            row,
                                            TQuantityValueMicrometer,
                                        ),
                                    ),
                                    processed_data_custom_document,
                                ),
                            ]
                        ),
                    )
                ],
            ),
        )



from allotropy.allotrope.models.adm.electrophoresis.benchling._2024._06.electrophoresis import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.electrophoresis.benchling._2024._06.electrophoresis import (
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.agilent_tapestation_analysis.agilent_tapestation_analysis_structure import (
    create_data,
)
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser


class AgilentTapestationAnalysisParser(VendorParser):
    @property
    def display_name(self) -> str:
        return "Agilent TapeStation Analysis"

    @property
    def release_state(self) -> ReleaseState:
        return ReleaseState.RECOMMENDED

    def to_allotrope(self, named_file_contents: NamedFileContents) -> Model:
        data = create_data(named_file_contents)
        return self._get_mapper(Mapper).map_model(data)



from collections.abc import Callable
from dataclasses import dataclass

from allotropy.allotrope.models.adm.electrophoresis.benchling._2024._06.electrophoresis import (
    CalculatedDataAggregateDocument,
    CalculatedDataDocumentItem,
    DataRegionAggregateDocument,
    DataRegionDocumentItem,
    DataSourceAggregateDocument,
    DataSourceDocumentItem,
    DataSystemDocument,
    DeviceControlAggregateDocument,
    DeviceControlDocumentItem,
    DeviceSystemDocument,
    ElectrophoresisAggregateDocument,
    ElectrophoresisDocumentItem,
    ErrorAggregateDocument,
    ErrorDocumentItem,
    MeasurementAggregateDocument,
    MeasurementDocumentItem,
    Model,
    PeakItem,
    PeakList,
    ProcessedDataAggregateDocument,
    ProcessedDataDocumentItem,
    SampleDocument,
    TQuantityValueModel,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueDegreeCelsius,
    TQuantityValuePercent,
    TQuantityValueRelativeFluorescenceUnit,
    TQuantityValueUnitless,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    JsonFloat,
    TDateTimeValue,
)
from allotropy.constants import ASM_CONVERTER_VERSION
from allotropy.parsers.utils.units import get_quantity_class
from allotropy.parsers.utils.values import assert_not_none, quantity_or_none


@dataclass(frozen=True)
class ProcessedDataFeature:
    identifier: str
    start: JsonFloat
    start_unit: str
    end: JsonFloat
    end_unit: str
    area: JsonFloat
    relative_area: JsonFloat
    position: JsonFloat | None = None
    position_unit: str | None = None
    height: JsonFloat | None = None
    relative_corrected_area: JsonFloat | None = None
    name: str | None = None
    comment: str | None = None


@dataclass(frozen=True)
class ProcessedData:
    peaks: list[ProcessedDataFeature]
    data_regions: list[ProcessedDataFeature]


@dataclass(frozen=True)
class DataSource:
    identifier: str
    feature: str


@dataclass(frozen=True)
class CalculatedDataItem:
    identifier: str
    name: str
    value: float
    unit: str
    data_sources: list[DataSource]


@dataclass(frozen=True)
class Error:
    error: str
    feature: str | None = None


@dataclass(frozen=True)
class Measurement:
    # Measurement metadata
    identifier: str
    measurement_time: str
    sample_identifier: str

    # Processed data
    processed_data: ProcessedData

    # Optional metadata
    description: str | None = None
    location_identifier: str | None = None

    # Optional settings
    compartment_temperature: float | None = None

    # Optional processed data
    calculated_data: list[CalculatedDataItem] | None = None

    # Errors
    errors: list[Error] | None = None


@dataclass(frozen=True)
class MeasurementGroup:
    measurements: list[Measurement]


@dataclass(frozen=True)
class Metadata:
    device_type: str
    device_identifier: str | None = None
    model_number: str | None = None
    software_name: str | None = None
    detection_type: str | None = None
    unc_path: str | None = None
    software_version: str | None = None
    equipment_serial_number: str | None = None
    product_manufacturer: str | None = None
    brand_name: str | None = None

    file_name: str | None = None
    data_system_instance_identifier: str | None = None

    analyst: str | None = None
    measurement_time: str | None = None
    analytical_method_identifier: str | None = None
    method_version: str | None = None
    experimental_data_identifier: str | None = None


@dataclass(frozen=True)
class Data:
    metadata: Metadata
    measurement_groups: list[MeasurementGroup]
    calculated_data: list[CalculatedDataItem] | None = None


class Mapper:
    MANIFEST = "http://purl.allotrope.org/manifests/electrophoresis/BENCHLING/2024/06/electrophoresis.manifest"

    def __init__(
        self, asm_converter_name: str, get_date_time: Callable[[str], TDateTimeValue]
    ) -> None:
        self.converter_name = asm_converter_name
        self.get_date_time = get_date_time

    def map_model(self, data: Data) -> Model:
        return Model(
            electrophoresis_aggregate_document=ElectrophoresisAggregateDocument(
                device_system_document=DeviceSystemDocument(
                    brand_name=data.metadata.brand_name,
                    product_manufacturer=data.metadata.product_manufacturer,
                    device_identifier=data.metadata.device_identifier,
                    equipment_serial_number=data.metadata.equipment_serial_number,
                ),
                data_system_document=DataSystemDocument(
                    data_system_instance_identifier=data.metadata.data_system_instance_identifier,
                    file_name=data.metadata.file_name,
                    ASM_converter_name=self.converter_name,
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                    software_name=data.metadata.software_name,
                    software_version=data.metadata.software_version,
                ),
                electrophoresis_document=[
                    self._get_technique_document(measurement_group, data.metadata)
                    for measurement_group in data.measurement_groups
                ],
                calculated_data_aggregate_document=self._get_calculated_data_aggregate_document(
                    data.calculated_data
                ),
            ),
            field_asm_manifest=self.MANIFEST,
        )

    def _get_technique_document(
        self, measurement_group: MeasurementGroup, metadata: Metadata
    ) -> ElectrophoresisDocumentItem:
        return ElectrophoresisDocumentItem(
            analyst=metadata.analyst,
            analytical_method_identifier=metadata.analytical_method_identifier,
            method_version=metadata.method_version,
            experimental_data_identifier=metadata.experimental_data_identifier,
            measurement_aggregate_document=MeasurementAggregateDocument(
                measurement_document=[
                    self._get_measurement_document_item(measurement, metadata)
                    for measurement in measurement_group.measurements
                ],
            ),
        )

    def _get_measurement_document_item(
        self, measurement: Measurement, metadata: Metadata
    ) -> MeasurementDocumentItem:
        return MeasurementDocumentItem(
            measurement_identifier=measurement.identifier,
            measurement_time=self.get_date_time(measurement.measurement_time),
            compartment_temperature=quantity_or_none(
                TQuantityValueDegreeCelsius,
                measurement.compartment_temperature,
            ),
            sample_document=self._get_sample_document(measurement),
            processed_data_aggregate_document=self._get_processed_data_aggregate_document(
                measurement.processed_data
            ),
            device_control_aggregate_document=DeviceControlAggregateDocument(
                device_control_document=[
                    DeviceControlDocumentItem(
                        device_type=metadata.device_type,
                        detection_type=metadata.detection_type,
                    ),
                ]
            ),
            calculated_data_aggregate_document=self._get_calculated_data_aggregate_document(
                measurement.calculated_data
            ),
            error_aggregate_document=self._get_error_aggregate_document(
                measurement.errors
            ),
        )

    def _get_sample_document(self, measurement: Measurement) -> SampleDocument:
        return SampleDocument(
            sample_identifier=measurement.sample_identifier,
            description=measurement.description,
            location_identifier=measurement.location_identifier,
        )

    def _get_processed_data_aggregate_document(
        self, data: ProcessedData
    ) -> ProcessedDataAggregateDocument:
        return ProcessedDataAggregateDocument(
            processed_data_document=[
                ProcessedDataDocumentItem(
                    peak_list=PeakList(
                        peak=[
                            PeakItem(
                                peak_identifier=peak.identifier,
                                peak_name=peak.name,
                                peak_height=quantity_or_none(
                                    TQuantityValueRelativeFluorescenceUnit, peak.height
                                ),
                                # TODO(nstender): figure out how to limit possible classes from get_quantity_class for typing.
                                peak_start=quantity_or_none(  # type: ignore[arg-type]
                                    assert_not_none(
                                        get_quantity_class(peak.start_unit)
                                    ),
                                    peak.start,
                                ),
                                peak_end=quantity_or_none(  # type: ignore[arg-type]
                                    assert_not_none(get_quantity_class(peak.end_unit)),
                                    peak.end,
                                ),
                                peak_position=quantity_or_none(  # type: ignore[arg-type]
                                    assert_not_none(
                                        get_quantity_class(peak.position_unit)
                                    ),
                                    peak.position,
                                ),
                                peak_area=quantity_or_none(
                                    TQuantityValueUnitless, peak.area
                                ),
                                relative_peak_area=quantity_or_none(
                                    TQuantityValuePercent, peak.relative_area
                                ),
                                relative_corrected_peak_area=quantity_or_none(
                                    TQuantityValuePercent, peak.relative_corrected_area
                                ),
                                comment=peak.comment,
                            )
                            for peak in data.peaks
                        ]
                    ),
                    data_region_aggregate_document=DataRegionAggregateDocument(
                        data_region_document=[
                            DataRegionDocumentItem(
                                region_identifier=data_region.identifier,
                                region_name=data_region.name,
                                # TODO(nstender): figure out how to limit possible classes from get_quantity_class for typing.
                                region_start=quantity_or_none(  # type: ignore[arg-type]
                                    assert_not_none(
                                        get_quantity_class(data_region.start_unit)
                                    ),
                                    data_region.start,
                                ),
                                region_end=quantity_or_none(  # type: ignore[arg-type]
                                    assert_not_none(
                                        get_quantity_class(data_region.end_unit)
                                    ),
                                    data_region.end,
                                ),
                                region_area=TQuantityValueUnitless(
                                    value=data_region.area
                                ),
                                relative_region_area=TQuantityValuePercent(
                                    value=data_region.relative_area
                                ),
                                comment=data_region.comment,
                            )
                            for data_region in data.data_regions
                        ]
                    )
                    if data.data_regions
                    else None,
                )
            ]
        )

    def _get_calculated_data_aggregate_document(
        self, calculated_data_items: list[CalculatedDataItem] | None
    ) -> CalculatedDataAggregateDocument | None:
        if not calculated_data_items:
            return None

        return CalculatedDataAggregateDocument(
            calculated_data_document=[
                CalculatedDataDocumentItem(
                    calculated_data_identifier=calculated_data_item.identifier,
                    calculated_data_name=calculated_data_item.name,
                    calculated_result=TQuantityValueModel(
                        value=calculated_data_item.value,
                        unit=calculated_data_item.unit,
                    ),
                    data_source_aggregate_document=DataSourceAggregateDocument(
                        data_source_document=[
                            DataSourceDocumentItem(
                                data_source_identifier=item.identifier,
                                data_source_feature=item.feature,
                            )
                            for item in calculated_data_item.data_sources
                        ]
                    ),
                )
                for calculated_data_item in calculated_data_items
            ]
        )

    def _get_error_aggregate_document(
        self, errors: list[Error] | None
    ) -> ErrorAggregateDocument | None:
        if not errors:
            return None

        return ErrorAggregateDocument(
            error_document=[
                ErrorDocumentItem(error=error.error, error_feature=error.feature)
                for error in errors
            ]
        )



from collections import defaultdict
from enum import Enum
from typing import cast

from allotropy.allotrope.models.adm.plate_reader.benchling._2023._09.plate_reader import (
    CalculatedDataAggregateDocument,
    CalculatedDataDocumentItem,
    ContainerType,
    DataSourceAggregateDocument,
    DataSourceDocumentItem,
    DataSystemDocument,
    DeviceControlDocument,
    DeviceSystemDocument,
    FluorescencePointDetectionDeviceControlAggregateDocument,
    FluorescencePointDetectionDeviceControlDocumentItem,
    FluorescencePointDetectionMeasurementDocumentItems,
    LuminescencePointDetectionDeviceControlAggregateDocument,
    LuminescencePointDetectionDeviceControlDocumentItem,
    LuminescencePointDetectionMeasurementDocumentItems,
    MeasurementAggregateDocument,
    Model,
    OpticalImagingMeasurementDocumentItems,
    PlateReaderAggregateDocument,
    PlateReaderDocumentItem,
    SampleDocument,
    UltravioletAbsorbancePointDetectionDeviceControlAggregateDocument,
    UltravioletAbsorbancePointDetectionDeviceControlDocumentItem,
    UltravioletAbsorbancePointDetectionMeasurementDocumentItems,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueDegreeCelsius,
    TQuantityValueMilliAbsorbanceUnit,
    TQuantityValueMillimeter,
    TQuantityValueNanometer,
    TQuantityValueNumber,
    TQuantityValueRelativeFluorescenceUnit,
    TQuantityValueRelativeLightUnit,
    TQuantityValueUnitless,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    TDateTimeValue,
)
from allotropy.constants import ASM_CONVERTER_VERSION
from allotropy.exceptions import AllotropeConversionError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.lines_reader import CsvReader, read_to_lines
from allotropy.parsers.perkin_elmer_envision.perkin_elmer_envision_structure import (
    CalculatedPlateInfo,
    Data,
    Plate,
    PlateMap,
    Result,
    ResultPlateInfo,
)
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.utils.values import quantity_or_none
from allotropy.parsers.vendor_parser import VendorParser


class ReadType(Enum):
    ABSORBANCE = "Absorbance"
    FLUORESCENCE = "Fluorescence"
    LUMINESCENCE = "Luminescence"


MeasurementDocumentItems = (
    OpticalImagingMeasurementDocumentItems
    | UltravioletAbsorbancePointDetectionMeasurementDocumentItems
    | FluorescencePointDetectionMeasurementDocumentItems
    | LuminescencePointDetectionMeasurementDocumentItems
)


DeviceControlAggregateDocument = (
    UltravioletAbsorbancePointDetectionDeviceControlAggregateDocument
    | FluorescencePointDetectionDeviceControlAggregateDocument
    | LuminescencePointDetectionDeviceControlAggregateDocument
)


class PerkinElmerEnvisionParser(VendorParser):
    @property
    def display_name(self) -> str:
        return "PerkinElmer Envision"

    @property
    def release_state(self) -> ReleaseState:
        return ReleaseState.RECOMMENDED

    def to_allotrope(self, named_file_contents: NamedFileContents) -> Model:
        lines = read_to_lines(named_file_contents)
        reader = CsvReader(lines)
        filename = named_file_contents.original_file_name
        return self._get_model(Data.create(reader), filename)

    def _get_model(self, data: Data, filename: str) -> Model:
        read_type = self._get_read_type(data)

        return Model(
            plate_reader_aggregate_document=PlateReaderAggregateDocument(
                plate_reader_document=self._get_plate_reader_document(data, read_type),
                data_system_document=DataSystemDocument(
                    file_name=filename,
                    software_name=data.software.software_name,
                    software_version=data.software.software_version,
                    ASM_converter_name=self.get_asm_converter_name(),
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                ),
                device_system_document=DeviceSystemDocument(
                    model_number="EnVision",
                    equipment_serial_number=data.instrument.serial_number,
                    device_identifier=data.instrument.nickname,
                ),
                calculated_data_aggregate_document=self._get_calculated_data_aggregate_document(
                    data, read_type
                ),
            ),
            field_asm_manifest="http://purl.allotrope.org/manifests/plate-reader/BENCHLING/2023/09/plate-reader.manifest",
        )

    def _get_read_type(self, data: Data) -> ReadType:
        patterns = {
            "ABS": ReadType.ABSORBANCE,
            "Absorbance": ReadType.ABSORBANCE,
            "LUM": ReadType.LUMINESCENCE,
            "Luminescence": ReadType.LUMINESCENCE,
            "Fluorescence": ReadType.FLUORESCENCE,
        }

        for key in patterns:
            if key in data.labels.label:
                return patterns[key]

        return (
            ReadType.FLUORESCENCE
        )  # TODO check if this is correct, this is the original behavior

    def _get_measurement_time(self, data: Data) -> TDateTimeValue:
        dates = [
            plate.plate_info.measurement_time
            for plate in data.plate_list.plates
            if plate.plate_info.measurement_time
        ]

        if dates:
            return self._get_date_time(min(dates))

        msg = "Unable to determine the measurement time."
        raise AllotropeConversionError(msg)

    def _get_device_control_aggregate_document(
        self,
        data: Data,
        result_plate_info: ResultPlateInfo,
        read_type: ReadType,
    ) -> DeviceControlAggregateDocument:
        ex_filter = data.labels.excitation_filter
        em_filter = data.labels.get_emission_filter(
            result_plate_info.emission_filter_id
        )

        if read_type == ReadType.LUMINESCENCE:
            return LuminescencePointDetectionDeviceControlAggregateDocument(
                device_control_document=[
                    LuminescencePointDetectionDeviceControlDocumentItem(
                        device_type="luminescence detector",
                        detector_distance_setting__plate_reader_=quantity_or_none(
                            TQuantityValueMillimeter, result_plate_info.measured_height
                        ),
                        number_of_averages=quantity_or_none(
                            TQuantityValueNumber, data.labels.number_of_flashes
                        ),
                        detector_gain_setting=data.labels.detector_gain_setting,
                        scan_position_setting__plate_reader_=data.labels.scan_position_setting,
                        detector_wavelength_setting=quantity_or_none(
                            TQuantityValueNanometer,
                            em_filter.wavelength if em_filter else None,
                        ),
                        detector_bandwidth_setting=quantity_or_none(
                            TQuantityValueNanometer,
                            em_filter.bandwidth if em_filter else None,
                        ),
                    )
                ]
            )
        elif read_type == ReadType.ABSORBANCE:
            return UltravioletAbsorbancePointDetectionDeviceControlAggregateDocument(
                device_control_document=[
                    UltravioletAbsorbancePointDetectionDeviceControlDocumentItem(
                        device_type="absorbance detector",
                        detector_distance_setting__plate_reader_=quantity_or_none(
                            TQuantityValueMillimeter, result_plate_info.measured_height
                        ),
                        number_of_averages=quantity_or_none(
                            TQuantityValueNumber, data.labels.number_of_flashes
                        ),
                        detector_gain_setting=data.labels.detector_gain_setting,
                        scan_position_setting__plate_reader_=data.labels.scan_position_setting,
                        detector_wavelength_setting=quantity_or_none(
                            TQuantityValueNanometer,
                            em_filter.wavelength if em_filter else None,
                        ),
                        detector_bandwidth_setting=quantity_or_none(
                            TQuantityValueNanometer,
                            em_filter.bandwidth if em_filter else None,
                        ),
                    )
                ]
            )
        else:  # read_type is FLUORESCENCE
            return FluorescencePointDetectionDeviceControlAggregateDocument(
                device_control_document=[
                    FluorescencePointDetectionDeviceControlDocumentItem(
                        device_type="fluorescence detector",
                        detector_distance_setting__plate_reader_=quantity_or_none(
                            TQuantityValueMillimeter, result_plate_info.measured_height
                        ),
                        number_of_averages=quantity_or_none(
                            TQuantityValueNumber, data.labels.number_of_flashes
                        ),
                        detector_gain_setting=data.labels.detector_gain_setting,
                        scan_position_setting__plate_reader_=data.labels.scan_position_setting,
                        detector_wavelength_setting=quantity_or_none(
                            TQuantityValueNanometer,
                            em_filter.wavelength if em_filter else None,
                        ),
                        detector_bandwidth_setting=quantity_or_none(
                            TQuantityValueNanometer,
                            em_filter.bandwidth if em_filter else None,
                        ),
                        excitation_wavelength_setting=quantity_or_none(
                            TQuantityValueNanometer,
                            ex_filter.wavelength if ex_filter else None,
                        ),
                        excitation_bandwidth_setting=quantity_or_none(
                            TQuantityValueNanometer,
                            ex_filter.bandwidth if ex_filter else None,
                        ),
                    )
                ]
            )

    def _get_measurement_document(
        self,
        plate: Plate,
        result: Result,
        p_map: PlateMap,
        device_control_document: list[DeviceControlDocument],
        read_type: ReadType,
    ) -> MeasurementDocumentItems:
        plate_barcode = plate.plate_info.barcode
        well_location = f"{result.col}{result.row}"
        sample_document = SampleDocument(
            sample_identifier=f"{plate_barcode} {well_location}",
            well_plate_identifier=plate_barcode,
            location_identifier=well_location,
            sample_role_type=p_map.get_sample_role_type(result.col, result.row).value,
        )
        compartment_temperature = quantity_or_none(
            TQuantityValueDegreeCelsius,
            plate.plate_info.chamber_temperature_at_start,
        )
        if read_type == ReadType.ABSORBANCE:
            return UltravioletAbsorbancePointDetectionMeasurementDocumentItems(
                measurement_identifier=result.uuid,
                sample_document=sample_document,
                device_control_aggregate_document=UltravioletAbsorbancePointDetectionDeviceControlAggregateDocument(
                    device_control_document=cast(
                        list[
                            UltravioletAbsorbancePointDetectionDeviceControlDocumentItem
                        ],
                        device_control_document,
                    ),
                ),
                absorbance=TQuantityValueMilliAbsorbanceUnit(value=result.value),
                compartment_temperature=compartment_temperature,
            )
        elif read_type == ReadType.LUMINESCENCE:
            return LuminescencePointDetectionMeasurementDocumentItems(
                measurement_identifier=result.uuid,
                sample_document=sample_document,
                device_control_aggregate_document=LuminescencePointDetectionDeviceControlAggregateDocument(
                    device_control_document=cast(
                        list[LuminescencePointDetectionDeviceControlDocumentItem],
                        device_control_document,
                    ),
                ),
                luminescence=TQuantityValueRelativeLightUnit(value=result.value),
                compartment_temperature=compartment_temperature,
            )
        else:  # read_type is FLUORESCENCE
            return FluorescencePointDetectionMeasurementDocumentItems(
                measurement_identifier=result.uuid,
                sample_document=sample_document,
                device_control_aggregate_document=FluorescencePointDetectionDeviceControlAggregateDocument(
                    device_control_document=cast(
                        list[FluorescencePointDetectionDeviceControlDocumentItem],
                        device_control_document,
                    ),
                ),
                fluorescence=TQuantityValueRelativeFluorescenceUnit(value=result.value),
                compartment_temperature=compartment_temperature,
            )

    def _get_plate_reader_document(
        self,
        data: Data,
        read_type: ReadType,
    ) -> list[PlateReaderDocumentItem]:
        items = []
        measurement_time = self._get_measurement_time(data)
        measurement_docs_dict = defaultdict(list)

        for plate in data.plate_list.plates:
            if isinstance(plate.plate_info, CalculatedPlateInfo):
                continue

            try:
                p_map = data.plate_maps[plate.plate_info.number]
            except KeyError as e:
                msg = f"Unable to find plate map for Plate {plate.plate_info.barcode}."
                raise AllotropeConversionError(msg) from e

            device_control_aggregate_document = (
                self._get_device_control_aggregate_document(
                    data, plate.plate_info, read_type
                )
            )

            for result in plate.result_list.results:
                measurement_docs_dict[
                    (plate.plate_info.number, result.col, result.row)
                ].append(
                    self._get_measurement_document(
                        plate,
                        result,
                        p_map,
                        cast(
                            list[DeviceControlDocument],
                            device_control_aggregate_document.device_control_document,
                        ),
                        read_type,
                    )
                )

        for well_location in sorted(measurement_docs_dict.keys()):
            items.append(
                PlateReaderDocumentItem(
                    measurement_aggregate_document=MeasurementAggregateDocument(
                        measurement_time=measurement_time,
                        plate_well_count=TQuantityValueNumber(
                            value=data.number_of_wells
                        ),
                        measurement_document=measurement_docs_dict[well_location],
                        analytical_method_identifier=data.basic_assay_info.protocol_id,
                        experimental_data_identifier=data.basic_assay_info.assay_id,
                        container_type=ContainerType.well_plate,
                    )
                )
            )

        return items

    def _get_calculated_data_aggregate_document(
        self,
        data: Data,
        read_type: ReadType,
    ) -> CalculatedDataAggregateDocument | None:
        calculated_documents = []

        for calculated_plate in data.plate_list.plates:
            if isinstance(calculated_plate.plate_info, ResultPlateInfo):
                continue

            source_result_lists = [
                source_plate.result_list.results
                for source_plate in calculated_plate.collect_result_plates(
                    data.plate_list
                )
            ]

            for calculated_result, *source_results in zip(
                calculated_plate.calculated_result_list.calculated_results,
                *source_result_lists,
                strict=True,
            ):
                calculated_documents.append(
                    CalculatedDataDocumentItem(
                        calculated_data_identifier=calculated_result.uuid,
                        calculated_data_name=calculated_plate.plate_info.name,
                        calculation_description=calculated_plate.plate_info.formula,
                        calculated_result=TQuantityValueUnitless(
                            value=calculated_result.value
                        ),
                        data_source_aggregate_document=DataSourceAggregateDocument(
                            data_source_document=[
                                DataSourceDocumentItem(
                                    data_source_identifier=source_result.uuid,
                                    data_source_feature=read_type.value,
                                )
                                for source_result in source_results
                            ]
                        ),
                    )
                )

        if not calculated_documents:
            return None

        return CalculatedDataAggregateDocument(
            calculated_data_document=calculated_documents,
        )



from enum import Enum


class XrVersion(str, Enum):
    _2_06 = "2.06"
    _2_04 = "2.04"


DATE_HEADER = "Sample date"

HEADINGS_TO_PARSER_HEADINGS = {
    "RunDate": DATE_HEADER,
    "Sample date/time": DATE_HEADER,
    # NOTE: data here is a typo in the Beckman software.
    "Sample data/time": DATE_HEADER,
    "Dilution": "Dilution factor",
    "Total cells / ml (x 10^6)": "Total cells/ml (x10^6)",
    "Average diameter (microns)": "Avg. diam. (microns)",
    "Average circularity": "Avg. circ.",
    "Total viable cells / ml (x 10^6)": "Viable cells/ml (x10^6)",
}

MODEL_RE = r"Vi-CELL XR (?P<version>\d{1,}\.\d{2,}(.\d{1,})?)"
DEFAULT_VERSION = XrVersion._2_06

DEFAULT_ANALYST = "Vi-Cell XR"
MODEL_NUMBER = "Vi-Cell XR"
SOFTWARE_NAME = "Vi-Cell XR"



""" Constants file for Roche Cedex HiRes Parser"""

# Instrument Software Details
CEDEX_SOFTWARE = "Cedex software"
MODEL_NUMBER = "Cedex HiRes"
BRAND_NAME = "Cedex"
PRODUCT_MANUFACTURER = "Roche"
DEVICE_TYPE = "cell counter"
DISPLAY_NAME = "Roche Cedex HiRes"
DETECTION_TYPE = "brightfield"

# Errors
MULTIPLE_SYSTEM_ERROR = "Error: Multiple values detected for system parameters. Expected a single unique value for:"
VALUE_ERROR = "Unable to find value for column"
UNSUPPORTED_FILE_FORMAT_ERROR = (
    "Unsupported file format. Expected xlsx or csv file. Actual: "
)



from __future__ import annotations

import pandas as pd

from allotropy.allotrope.schema_mappers.adm.cell_counting.benchling._2023._11.cell_counting import (
    Data,
    Measurement,
    MeasurementGroup,
    Metadata,
)
from allotropy.parsers.beckman_vi_cell_blu.constants import (
    DEFAULT_ANALYST,
    DEFAULT_MODEL_NUMBER,
    VICELL_BLU_SOFTWARE_NAME,
)
from allotropy.parsers.utils.pandas import map_rows, SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str


def _create_measurement_group(data: SeriesData) -> MeasurementGroup:
    total_cell_count = data.get(float, "Cell count")
    total_cell_count = (
        total_cell_count if total_cell_count is None else round(total_cell_count)
    )
    viable_cell_count = data.get(float, "Viable cells")
    viable_cell_count = (
        viable_cell_count if viable_cell_count is None else round(viable_cell_count)
    )

    return MeasurementGroup(
        measurements=[
            Measurement(
                measurement_identifier=random_uuid_str(),
                timestamp=data[str, "Analysis date/time"],
                sample_identifier=data[str, "Sample ID"],
                cell_type_processing_method=data.get(str, "Cell type"),
                minimum_cell_diameter_setting=data.get(float, "Minimum Diameter (μm)"),
                maximum_cell_diameter_setting=data.get(float, "Maximum Diameter (μm)"),
                cell_density_dilution_factor=data.get(float, "Dilution"),
                viability=data[float, "Viability (%)"],
                viable_cell_density=data[float, "Viable (x10^6) cells/mL"],
                total_cell_count=total_cell_count,
                total_cell_density=data.get(float, "Total (x10^6) cells/mL"),
                average_total_cell_diameter=data.get(float, "Average diameter (μm)"),
                average_live_cell_diameter=data.get(
                    float, "Average viable diameter (μm)"
                ),
                viable_cell_count=viable_cell_count,
                average_total_cell_circularity=data.get(float, "Average circularity"),
                average_viable_cell_circularity=data.get(
                    float, "Average viable circularity"
                ),
                analyst=data.get(str, "Analysis by", DEFAULT_ANALYST),
            )
        ]
    )


def create_data(data: pd.DataFrame, file_name: str) -> Data:
    metadata = Metadata(
        device_type="brightfield imager (cell counter)",
        detection_type="brightfield",
        model_number=DEFAULT_MODEL_NUMBER,
        software_name=VICELL_BLU_SOFTWARE_NAME,
        file_name=file_name,
    )
    return Data(metadata, map_rows(data, _create_measurement_group))



from __future__ import annotations

import numpy as np

from allotropy.allotrope.models.adm.plate_reader.benchling._2023._09.plate_reader import (
    Model,
)
from allotropy.allotrope.pandas_util import read_csv
from allotropy.allotrope.schema_mappers.adm.plate_reader.benchling._2023._09.plate_reader import (
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.unchained_labs_lunatic.unchained_labs_lunatic_structure import (
    create_data,
)
from allotropy.parsers.vendor_parser import VendorParser


class UnchainedLabsLunaticParser(VendorParser):
    @property
    def display_name(self) -> str:
        return "Unchained Labs Lunatic"

    @property
    def release_state(self) -> ReleaseState:
        return ReleaseState.RECOMMENDED

    def to_allotrope(self, named_file_contents: NamedFileContents) -> Model:
        raw_contents = named_file_contents.contents
        contents = read_csv(filepath_or_buffer=raw_contents).replace(np.nan, None)
        data = create_data(contents, named_file_contents.original_file_name)
        return self._get_mapper(Mapper).map_model(data)



from allotropy.allotrope.models.adm.plate_reader.benchling._2023._09.plate_reader import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.plate_reader.benchling._2023._09.plate_reader import (
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.agilent_gen5.agilent_gen5_structure import create_data
from allotropy.parsers.agilent_gen5.section_reader import SectionLinesReader
from allotropy.parsers.lines_reader import read_to_lines
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser


class AgilentGen5Parser(VendorParser):
    @property
    def display_name(self) -> str:
        return "Agilent Gen5"

    @property
    def release_state(self) -> ReleaseState:
        return ReleaseState.RECOMMENDED

    def to_allotrope(self, named_file_contents: NamedFileContents) -> Model:
        lines = read_to_lines(named_file_contents)
        data = create_data(
            SectionLinesReader(lines), named_file_contents.original_file_name
        )
        return self._get_mapper(Mapper).map_model(data)



from __future__ import annotations

from collections.abc import Callable, Iterable
from enum import Enum
import re
from typing import Any, Literal, overload, TypeVar

import pandas as pd

from allotropy.allotrope.models.shared.definitions.definitions import (
    InvalidJsonFloat,
)
from allotropy.exceptions import AllotropeConversionError, AllotropyParserError
from allotropy.parsers.utils.iterables import get_first_not_none
from allotropy.parsers.utils.values import (
    assert_not_none,
    str_to_bool,
)

MapType = TypeVar("MapType")


def map_rows(
    data_frame: pd.DataFrame, func: Callable[[SeriesData], MapType]
) -> list[MapType]:
    def run_with_data(series: pd.Series[str]) -> MapType:
        return func(SeriesData(series))

    # pandas can't find a matching overload for this, but it works and returns the correct type...
    return list(data_frame.apply(run_with_data, axis="columns"))  # type: ignore[call-overload]


def rm_df_columns(data: pd.DataFrame, pattern: str) -> pd.DataFrame:
    return data.drop(
        columns=[column for column in data.columns if re.match(pattern, column)]
    )


def df_to_series(
    df: pd.DataFrame, msg: str, index: int | None = None
) -> pd.Series[Any]:
    n_rows, _ = df.shape
    if index is None and n_rows == 1:
        index = 0
    if index is None or index >= n_rows:
        raise AllotropyParserError(msg)
    return pd.Series(df.iloc[index], index=df.columns)


def df_to_series_data(df: pd.DataFrame, msg: str) -> SeriesData:
    return SeriesData(df_to_series(df, msg))


def assert_df_column(df: pd.DataFrame, column: str) -> pd.Series[Any]:
    df_column = df.get(column)
    if df_column is None:
        msg = f"Unable to find column '{column}'."
        raise AllotropeConversionError(msg)
    return pd.Series(df_column)


def assert_not_empty_df(df: pd.DataFrame, msg: str) -> pd.DataFrame:
    if df.empty:
        raise AllotropeConversionError(msg)
    return df


T = TypeVar("T", bool, float, int, str)
# Need to use this instead of type[T] to get mypy to realize primitive can be called to return T
Type_ = Callable[..., T]
KeyOrKeys = Iterable[str]
TypeAndKey = tuple[Type_[T], KeyOrKeys]
TypeAndKeyAndMsg = tuple[Type_[T], KeyOrKeys, str]
ValidateRaw = Callable[[Any], bool] | None


class SeriesData:
    class ValidateRawMode(Enum):
        # Return None for key is raw value is None or np.isna
        NOT_NAN = "NOT_NAN"

    NOT_NAN = ValidateRawMode.NOT_NAN

    @staticmethod
    def _validate_raw(v: Any, mode: ValidateRawMode | None) -> Any:
        if mode is SeriesData.ValidateRawMode.NOT_NAN:
            return None if (v is None or pd.isna(v)) else v
        return v

    def __init__(self, series: pd.Series[Any]) -> None:
        self.series = series

    def __getitem__(self, type_and_key: TypeAndKey[T] | TypeAndKeyAndMsg[T]) -> T:
        """
        Get a value of the specified type with the specified key, raising an error if the
        key is not found, or if the value cannot be converted to the type.
        If a third argument is provided, it is an error message to provide if they key is not found.

        value: float = series_data[float, key]
        value: str = series_data[str, key, f"Failed to find {key} in my data"]

        Parameters:
        type (str, int, float, bool): The datatype to return.
        key (str | Iterable[str]): The key (or iterable of keys) to use to lookup.
        msg (str | None): The message to give as an error if lookup or conversion fails.

        Returns:
        type: A value of the type provided.

        Raises
        AllotropeConversionError: If the lookup or conversion to type fails.
        """
        if len(type_and_key) == 2:  # noqa: PLR2004
            type_, key = type_and_key
            msg = None
        elif len(type_and_key) == 3:  # noqa: PLR2004
            type_, key, msg = type_and_key
        return assert_not_none(self.get(type_, key), str(key), msg=msg)

    # This overload tells typing that if default is "None" then get might return None
    @overload
    def get(
        self,
        type_: Type_[T],
        key: KeyOrKeys,
        default: Literal[None] = None,
        validate: ValidateRawMode | None = None,
    ) -> T | None:
        ...

    # This overload tells typing that if default matches T, get will return T
    @overload
    def get(
        self,
        type_: Type_[float],
        key: KeyOrKeys,
        default: InvalidJsonFloat,
        validate: ValidateRawMode | None = None,
    ) -> float | InvalidJsonFloat:
        ...

    # This overload tells typing that if default matches T, get will return T
    @overload
    def get(
        self,
        type_: Type_[T],
        key: KeyOrKeys,
        default: T,
        validate: ValidateRawMode | None = None,
    ) -> T:
        ...

    def get(
        self,
        type_: Type_[T],
        key: KeyOrKeys,
        default: T | InvalidJsonFloat | None = None,
        validate: ValidateRawMode | None = None,
    ) -> T | InvalidJsonFloat | None:
        """
        Get a value of the specified type with the specified key, returning a default value if the
        key is not found, or if the value cannot be converted to the type.
        If validate is provided, it will be used to validate the raw value from the series, returning None
        if the value fails validation.

        Parameters:
        type (str, int, float, bool): The datatype to return.
        key (str | Iterable[str]): The key (or iterable of keys) to use to lookup.
        default (type | InvalidJsonFloat | None): The value to return if lookup or conversion fails (default=None).
        validate (ValidateRawMode): The method to use for validating raw value. Defaults to (value is not None).

        Returns:
        type: A value of the type provided or default value.
        """
        if not isinstance(key, str):
            return get_first_not_none(
                lambda k: self.get(type_, k, validate=validate), key
            )
        raw_value = self._validate_raw(self.series.get(key), validate)
        try:
            # bool needs special handling to convert
            if type_ is bool:
                raw_value = (
                    None
                    if raw_value is None
                    else ("true" if str_to_bool(str(raw_value)) else "")
                )
            if type_ is float and isinstance(raw_value, str) and "%" in raw_value:
                raw_value = raw_value.strip("%")
            value = None if raw_value is None else type_(raw_value)
        except ValueError:
            value = None
        return default if value is None else value



""" Constants file for Thermo Fisher Qubit Flex Parser"""

# Instrument constants
SOFTWARE_NAME = "Qubit Flex software"
MODEL_NUMBER = "Qubit Flex"
PRODUCT_MANUFACTURER = "Thermo Fisher Scientific"
BRAND_NAME = "Qubit"
DEVICE_TYPE = "fluorescence detector"
DISPLAY_NAME = "Thermo Fisher Qubit Flex"

# Error constants
# Value to use for unsupported file format for the given file for the Adapter
UNSUPPORTED_FILE_FORMAT_ERROR = (
    "Unsupported file format. Expected xlsx or csv file. Actual: "
)
VALUE_ERROR = "Unable to find value for column"



#!/usr/bin/env python3
from datetime import datetime
import os
from pathlib import Path
import subprocess

import click
from pytz import timezone
import semantic_version  # type: ignore

from allotropy.__about__ import __version__
from allotropy.allotrope.schema_parser.path_util import ALLOTROPY_DIR, ROOT_DIR


def _update_changelog(version: str) -> str:
    changelog_file = Path(ROOT_DIR, "CHANGELOG.md")
    with open(changelog_file) as f:
        contents = f.readlines()

    body = ""
    with open(changelog_file, "w") as f:
        editing = None
        current_contents: list[str] = []
        for line in contents:
            if line.startswith("## "):
                if editing:
                    # Add last section to body and write body
                    if "".join(current_contents[1:]).strip():
                        body += "".join(
                            f"- {c.strip()}\n"
                            if not (c.startswith("-") or c.startswith("\n"))
                            else c
                            for c in current_contents
                            if c != "\n"
                        )
                    f.write(body)

                    # Write start of new section
                    f.write("\n")

                    editing = False
                elif editing is None:
                    # Start writing current changelog entry, skipping empty sections.
                    f.write(
                        """## [Unreleased]\n\n### Added\n\n### Fixed\n\n### Changed\n\n### Deprecated\n\n### Removed\n\n### Security\n\n"""
                    )
                    f.write(
                        f"## [{version}] - {datetime.now(timezone('EST')).strftime('%Y-%m-%d')}\n"
                    )
                    editing = True
                    continue

            if editing:
                if line.startswith("### "):
                    if "".join(current_contents[1:]).strip():
                        body += "".join(
                            f"- {c.strip()}\n"
                            if not (c.startswith("-") or c.startswith("\n"))
                            else c
                            for c in current_contents
                            if c != "\n"
                        )
                    current_contents = [f"\n{line}\n"]
                else:
                    current_contents.append(line)
                continue

            f.write(line)

    return body


def _write_version_file(version: str) -> None:
    with open(Path(ALLOTROPY_DIR, "__about__.py"), "w") as f:
        f.write(f'__version__ = "{version}"\n')


def _make_pr(version: str, body: str) -> None:
    try:
        subprocess.run(
            ["gh", "--help"],
            check=True,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
    except FileNotFoundError:
        print(
            "gh not installed - cannot create PR automatically. Try: 'brew install gh'"
        )
        return

    print("Making commit...")
    subprocess.run(
        [
            "git",
            "commit",
            "-am",
            f"Update allotropy version to {version}",
        ],
        check=True,
    )
    print("Pushing commit...")
    subprocess.run(["git", "push"], check=True)

    print("Creating PR...")
    filename = ".temp_pr_description.txt"
    with open(filename, "w") as f:
        f.write(body)

    try:
        subprocess.run(
            [
                "gh",
                "pr",
                "create",
                "--title",
                f"chore: Update allotropy version to {version}",
                "--body-file",
                filename,
            ],
            check=True,
        )
    finally:
        os.remove(filename)


@click.command()
@click.option(
    "--version", "-v", help="Version to update to, defaults to next patch version"
)
@click.option(
    "--skip_pr", is_flag=True, default=False, help="Whether to make PR in script"
)
def _update_version(
    version: str | None = None, skip_pr: bool = False  # noqa: FBT001, FBT002
) -> None:
    """Update allotropy version."""
    if version:
        semver = semantic_version.Version(version)
    if not version:
        semver = semantic_version.Version(__version__)
        semver.patch += 1

    version = str(semver)

    print("Updating version file and CHANGELOG...")
    _write_version_file(version)
    body = _update_changelog(version)
    if not skip_pr:
        _make_pr(version, body)


if __name__ == "__main__":
    _update_version()



from collections.abc import Callable
from dataclasses import dataclass
from typing import TypeVar

from allotropy.allotrope.models.adm.pcr.benchling._2023._09.qpcr import (
    BaselineCorrectedReporterDataCube,
    CalculatedDataDocumentItem,
    ContainerType,
    DataProcessingDocument,
    DataSourceAggregateDocument,
    DataSourceDocumentItem,
    DataSystemDocument,
    DeviceControlAggregateDocument,
    DeviceControlDocumentItem,
    DeviceSystemDocument,
    ExperimentType,
    MeasurementAggregateDocument,
    MeasurementDocumentItem,
    MeltingCurveDataCube,
    Model,
    NormalizedReporterDataCube,
    PassiveReferenceDyeDataCube,
    ProcessedDataAggregateDocument,
    ProcessedDataDocumentItem,
    QPCRAggregateDocument,
    QPCRDocumentItem,
    ReporterDyeDataCube,
    SampleDocument,
    TCalculatedDataAggregateDocument,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TNullableQuantityValueUnitless,
    TQuantityValueNumber,
    TQuantityValueUnitless,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    FieldComponentDatatype,
    InvalidJsonFloat,
    TDatacubeComponent,
    TDatacubeData,
    TDatacubeStructure,
    TDateTimeValue,
)
from allotropy.constants import ASM_CONVERTER_VERSION
from allotropy.parsers.utils.iterables import get_first_not_none
from allotropy.parsers.utils.values import assert_not_none, quantity_or_none


@dataclass
class DataCubeComponent:
    type_: FieldComponentDatatype
    concept: str
    unit: str


@dataclass
class DataCube:
    label: str
    structure_dimensions: list[DataCubeComponent]
    structure_measures: list[DataCubeComponent]
    dimensions: list[list[float]]
    measures: list[list[float | None]]


@dataclass
class DataSource:
    identifier: str
    feature: str


@dataclass
class CalculatedDataItem:
    identifier: str
    name: str
    value: float
    unit: str
    data_sources: list[DataSource]


@dataclass
class CalculatedData:
    items: list[CalculatedDataItem]
    reference_dna_description: str | None = None
    reference_sample_description: str | None = None


@dataclass
class ProcessedData:
    cycle_threshold_value_setting: float
    automatic_cycle_threshold_enabled_setting: bool | None = None
    automatic_baseline_determination_enabled_setting: bool | None = None
    genotyping_determination_method_setting: float | None = None
    genotyping_determination_result: str | None = None
    cycle_threshold_result: float | None = None
    normalized_reporter_result: float | None = None
    baseline_corrected_reporter_result: float | None = None
    data_cubes: list[DataCube] | None = None


@dataclass
class Measurement:
    # Measurement metadata
    identifier: str
    timestamp: str
    target_identifier: str
    sample_identifier: str

    # Settings
    pcr_detection_chemistry: str

    # Optional measurement metadata
    sample_role_type: str | None = None
    well_location_identifier: str | None = None
    well_plate_identifier: str | None = None

    # Optional settings
    total_cycle_number_setting: float | None = None
    reporter_dye_setting: str | None = None
    quencher_dye_setting: str | None = None
    passive_reference_dye_setting: str | None = None

    # Processed data
    processed_data: ProcessedData | None = None
    data_cubes: list[DataCube] | None = None


@dataclass
class MeasurementGroup:
    measurements: list[Measurement]
    plate_well_count: int | InvalidJsonFloat
    analyst: str | None = None
    experimental_data_identifier: str | None = None


@dataclass
class Metadata:
    device_identifier: str
    device_type: str
    device_serial_number: str
    model_number: str
    software_name: str
    software_version: str
    data_system_instance_identifier: str
    file_name: str
    unc_path: str
    experiment_type: ExperimentType
    container_type: ContainerType
    measurement_method_identifier: str


@dataclass
class Data:
    metadata: Metadata
    measurement_groups: list[MeasurementGroup]
    calculated_data: CalculatedData


CubeClass = TypeVar("CubeClass")


class Mapper:
    MANIFEST = "http://purl.allotrope.org/manifests/cell-counting/BENCHLING/2023/11/cell-counting.manifest"

    def __init__(
        self, asm_converter_name: str, get_date_time: Callable[[str], TDateTimeValue]
    ) -> None:
        self.converter_name = asm_converter_name
        self.get_date_time = get_date_time

    def map_model(self, data: Data) -> Model:
        return Model(
            qPCR_aggregate_document=QPCRAggregateDocument(
                device_system_document=DeviceSystemDocument(
                    device_identifier=data.metadata.device_identifier,
                    model_number=data.metadata.model_number,
                    device_serial_number=data.metadata.device_serial_number,
                ),
                data_system_document=DataSystemDocument(
                    data_system_instance_identifier=data.metadata.data_system_instance_identifier,
                    file_name=data.metadata.file_name,
                    UNC_path=data.metadata.unc_path,
                    software_name=data.metadata.software_name,
                    software_version=data.metadata.software_version,
                    ASM_converter_name=self.converter_name,
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                ),
                qPCR_document=[
                    self._get_technique_document(measurement_group, data.metadata)
                    for measurement_group in data.measurement_groups
                ],
                calculated_data_aggregate_document=self._get_calculated_data_aggregate_document(
                    data
                ),
            )
        )

    def _get_technique_document(
        self, measurement_group: MeasurementGroup, metadata: Metadata
    ) -> QPCRDocumentItem:
        return QPCRDocumentItem(
            analyst=measurement_group.analyst,
            measurement_aggregate_document=MeasurementAggregateDocument(
                experimental_data_identifier=measurement_group.experimental_data_identifier,
                experiment_type=metadata.experiment_type,
                container_type=metadata.container_type,
                plate_well_count=TQuantityValueNumber(
                    value=measurement_group.plate_well_count
                ),
                measurement_document=[
                    self._get_measurement_document_item(measurement, metadata)
                    for measurement in measurement_group.measurements
                ],
            ),
        )

    def _get_measurement_document_item(
        self,
        measurement: Measurement,
        metadata: Metadata,
    ) -> MeasurementDocumentItem:
        return MeasurementDocumentItem(
            measurement_identifier=measurement.identifier,
            measurement_time=self.get_date_time(measurement.timestamp),
            target_DNA_description=measurement.target_identifier,
            sample_document=self._get_sample_document(measurement),
            device_control_aggregate_document=DeviceControlAggregateDocument(
                device_control_document=[
                    DeviceControlDocumentItem(
                        device_type=metadata.device_type,
                        measurement_method_identifier=metadata.measurement_method_identifier,
                        total_cycle_number_setting=quantity_or_none(
                            TQuantityValueNumber, measurement.total_cycle_number_setting
                        ),
                        PCR_detection_chemistry=measurement.pcr_detection_chemistry,
                        reporter_dye_setting=measurement.reporter_dye_setting,
                        quencher_dye_setting=measurement.quencher_dye_setting,
                        passive_reference_dye_setting=measurement.passive_reference_dye_setting,
                    )
                ],
            ),
            processed_data_aggregate_document=assert_not_none(
                self._get_processed_data_aggregate_document(measurement.processed_data)
            ),
            reporter_dye_data_cube=self._get_data_cube(
                ReporterDyeDataCube, "reporter dye", measurement.data_cubes
            ),
            passive_reference_dye_data_cube=self._get_data_cube(
                PassiveReferenceDyeDataCube,
                "passive reference dye",
                measurement.data_cubes,
            ),
            melting_curve_data_cube=self._get_data_cube(
                MeltingCurveDataCube, "melting curve", measurement.data_cubes
            ),
        )

    def _get_sample_document(self, measurement: Measurement) -> SampleDocument:
        return SampleDocument(
            sample_identifier=measurement.sample_identifier,
            sample_role_type=measurement.sample_role_type,
            well_location_identifier=measurement.well_location_identifier,
            well_plate_identifier=measurement.well_plate_identifier,
        )

    def _get_processed_data_aggregate_document(
        self, data: ProcessedData | None
    ) -> ProcessedDataAggregateDocument | None:
        if not data:
            return None
        return ProcessedDataAggregateDocument(
            processed_data_document=[
                ProcessedDataDocumentItem(
                    data_processing_document=DataProcessingDocument(
                        automatic_cycle_threshold_enabled_setting=data.automatic_baseline_determination_enabled_setting,
                        cycle_threshold_value_setting=TQuantityValueUnitless(
                            value=data.cycle_threshold_value_setting,
                        ),
                        automatic_baseline_determination_enabled_setting=data.automatic_baseline_determination_enabled_setting,
                        genotyping_determination_method_setting=quantity_or_none(
                            TQuantityValueUnitless,
                            data.genotyping_determination_method_setting,
                        ),
                    ),
                    cycle_threshold_result=TNullableQuantityValueUnitless(
                        value=data.cycle_threshold_result,
                    ),
                    normalized_reporter_result=quantity_or_none(
                        TQuantityValueUnitless, data.normalized_reporter_result
                    ),
                    baseline_corrected_reporter_result=quantity_or_none(
                        TQuantityValueUnitless,
                        data.baseline_corrected_reporter_result,
                    ),
                    genotyping_determination_result=data.genotyping_determination_result,
                    normalized_reporter_data_cube=self._get_data_cube(
                        NormalizedReporterDataCube,
                        "normalized reporter",
                        data.data_cubes,
                    ),
                    baseline_corrected_reporter_data_cube=self._get_data_cube(
                        BaselineCorrectedReporterDataCube,
                        "baseline corrected reporter",
                        data.data_cubes,
                    ),
                )
            ]
        )

    def _get_calculated_data_aggregate_document(
        self, data: Data
    ) -> TCalculatedDataAggregateDocument | None:
        if not data.calculated_data:
            return None

        if data.calculated_data.reference_sample_description:
            data_processing_document = DataProcessingDocument(
                reference_DNA_description=data.calculated_data.reference_dna_description,
                reference_sample_description=data.calculated_data.reference_sample_description,
            )
        else:
            data_processing_document = None

        return TCalculatedDataAggregateDocument(
            calculated_data_document=[
                CalculatedDataDocumentItem(
                    calculated_data_identifier=calc_doc.identifier,
                    data_source_aggregate_document=DataSourceAggregateDocument(
                        data_source_document=[
                            DataSourceDocumentItem(
                                data_source_identifier=data_source.identifier,
                                data_source_feature=data_source.feature,
                            )
                            for data_source in calc_doc.data_sources
                        ],
                    ),
                    data_processing_document=data_processing_document,
                    calculated_data_name=calc_doc.name,
                    calculated_data_description=None,
                    calculated_datum=TQuantityValueUnitless(
                        value=calc_doc.value, unit=calc_doc.unit
                    ),
                )
                for calc_doc in data.calculated_data.items
            ],
        )

    def _get_data_cube(
        self,
        cube_class: Callable[..., CubeClass],
        label: str,
        data_cubes: list[DataCube] | None,
    ) -> CubeClass | None:
        if not (
            data_cube := get_first_not_none(
                lambda cube: cube if cube.label == label else None,
                data_cubes or [],
            )
        ):
            return None

        return cube_class(
            label=data_cube.label,
            cube_structure=TDatacubeStructure(
                dimensions=[
                    TDatacubeComponent(
                        field_componentDatatype=component.type_,
                        concept=component.concept,
                        unit=component.unit,
                    )
                    for component in data_cube.structure_dimensions
                ],
                measures=[
                    TDatacubeComponent(
                        field_componentDatatype=component.type_,
                        concept=component.concept,
                        unit=component.unit,
                    )
                    for component in data_cube.structure_measures
                ],
            ),
            data=TDatacubeData(
                dimensions=data_cube.dimensions, measures=data_cube.measures  # type: ignore[arg-type]
            ),
        )
